# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

from jaxley.utils.cell_utils import local_index_of_loc

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import numpy as np
import pytest

import jaxley as jx
from jaxley.connect import (
    connect,
    connectivity_matrix_connect,
    fully_connect,
    sparse_connect,
)
from jaxley.synapses import IonotropicSynapse, TestSynapse


def test_connect(SimpleBranch, SimpleCell, SimpleNet):
    branch = SimpleBranch(4)
    cell = SimpleCell(3, 4)
    net1 = SimpleNet(4, 3, 8)
    net2 = SimpleNet(4, 3, 8)

    cell1_net1 = net1[0, 0, 0]
    cell2_net1 = net1[1, 0, 0]
    cell1_net2 = net2[0, 0, 0]

    ### test connect single
    # test raise if not part of net
    with pytest.raises(AssertionError):
        connect(cell[0, 0], cell[0, 1], TestSynapse())  # should raise
    with pytest.raises(AssertionError):
        connect(branch[0], branch[1], TestSynapse())  # should raise
    with pytest.raises(AssertionError):
        connect(cell[0, 0], branch[0], TestSynapse())  # should raise

    # test raise if not part of same net
    connect(cell1_net1, cell2_net1, TestSynapse())
    with pytest.raises(AssertionError):
        connect(cell1_net1, cell1_net2, TestSynapse())  # should raise

    ### test connect multiple
    # test connect multiple with single synapse
    connect(net2[1, 0], net2[2, 0], TestSynapse())

    # test after all connections are made, to catch "overwritten" connections
    get_comps = lambda locs: [
        local_index_of_loc(loc, 0, net2.ncomp_per_branch) for loc in locs
    ]

    # check if all connections are made correctly
    first_set_edges = net2.edges.iloc[:8]
    nodes = net2.nodes.set_index("global_comp_index")
    cols = ["pre_global_comp_index", "post_global_comp_index"]
    comp_inds = nodes.loc[first_set_edges[cols].to_numpy().flatten()]
    branch_inds = comp_inds["global_branch_index"].to_numpy().reshape(-1, 2)
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    assert np.all(branch_inds == (3, 6))
    assert (cell_inds == (1, 2)).all()
    assert (
        get_comps(first_set_edges["pre_locs"])
        == get_comps(first_set_edges["post_locs"])
        == [0, 1, 2, 3, 4, 5, 6, 7]
    )
    assert (first_set_edges["type"] == "TestSynapse").all()


def test_fully_connect():
    comp = jx.Compartment()
    branch = jx.Branch([comp for _ in range(8)])
    cell = jx.Cell([branch for _ in range(3)], parents=np.array([-1, 0, 0]))
    net = jx.Network([cell for _ in range(4 * 4)])

    _ = np.random.seed(0)
    for i in range(4):
        fully_connect(net[i], net[4:8], TestSynapse())

    fully_connect(net[8:12], net[12:16], TestSynapse())

    assert all(
        net.edges.post_global_comp_index
        == [
            108,
            135,
            165,
            168,
            99,
            123,
            151,
            177,
            115,
            141,
            162,
            172,
            119,
            126,
            156,
            169,
            294,
            329,
            345,
            379,
            295,
            317,
            356,
            365,
            311,
            325,
            355,
            375,
            302,
            320,
            352,
            375,
        ]
    )


def test_sparse_connect(SimpleNet):
    net = SimpleNet(4 * 4, 4, 4)

    _ = np.random.seed(0)
    for i in range(4):
        sparse_connect(net[i], net[4:8], TestSynapse(), p=0.5)

    sparse_connect(net[8:12], net[12:], TestSynapse(), p=0.5)

    assert all(
        [
            63,
            59,
            65,
            86,
            80,
            58,
            92,
            85,
            168,
            145,
            189,
            153,
            180,
            190,
            184,
            163,
            159,
            179,
            182,
        ]
    )


def test_connectivity_matrix_connect(SimpleNet):
    net = SimpleNet(4 * 4, 3, 8)

    _ = np.random.seed(0)
    n_by_n_adjacency_matrix = np.array(
        [[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [1, 0, 0, 0]], dtype=bool
    )
    incides_of_connected_cells = np.stack(np.where(n_by_n_adjacency_matrix)).T
    incides_of_connected_cells[:, 1] += 4

    connectivity_matrix_connect(
        net[:4], net[4:8], TestSynapse(), n_by_n_adjacency_matrix
    )
    assert len(net.edges.index) == 4
    nodes = net.nodes.set_index("global_comp_index")
    cols = ["pre_global_comp_index", "post_global_comp_index"]
    comp_inds = nodes.loc[net.edges[cols].to_numpy().flatten()]
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    assert np.all(cell_inds == incides_of_connected_cells)

    m_by_n_adjacency_matrix = np.array(
        [[0, 1, 1, 0], [0, 0, 1, 1], [0, 0, 0, 1]], dtype=bool
    )
    incides_of_connected_cells = np.stack(np.where(m_by_n_adjacency_matrix)).T

    net = SimpleNet(4 * 4, 3, 8)
    with pytest.raises(AssertionError):
        connectivity_matrix_connect(
            net[:4], net[:4], TestSynapse(), m_by_n_adjacency_matrix
        )  # should raise

    connectivity_matrix_connect(
        net[:3], net[:4], TestSynapse(), m_by_n_adjacency_matrix
    )
    assert len(net.edges.index) == 5
    nodes = net.nodes.set_index("global_comp_index")
    cols = ["pre_global_comp_index", "post_global_comp_index"]
    comp_inds = nodes.loc[net.edges[cols].to_numpy().flatten()]
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    assert np.all(cell_inds == incides_of_connected_cells)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")
from copy import copy

import jax.numpy as jnp
import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH, K, Na
from jaxley.connect import connect, fully_connect
from jaxley.synapses import IonotropicSynapse, TestSynapse
from jaxley.utils.cell_utils import params_to_pstate


def test_make_trainable(SimpleCell):
    """Test make_trainable."""
    cell = SimpleCell(4, 4)
    cell.insert(HH())

    cell.branch(0).loc(0.0).set("length", 12.0)
    cell.branch(1).loc(1.0).set("HH_gNa", 0.2)
    assert cell.num_trainable_params == 0

    cell.branch([0, 1]).make_trainable("radius", 1.0)
    assert cell.num_trainable_params == 2
    cell.branch([0, 1]).make_trainable("length")
    cell.branch([0, 1]).make_trainable("axial_resistivity", [600.0, 700.0])
    cell.branch([0, 1]).make_trainable("HH_gNa")

    cell.get_parameters()


def test_delete_trainables(SimpleCell):
    """Test make_trainable."""
    cell = SimpleCell(4, 4)

    cell.branch(0).loc(0.0).make_trainable("length", 12.0)
    assert cell.num_trainable_params == 1

    cell.delete_trainables()
    cell.branch(0).loc(0.0).make_trainable("length", 12.0)
    assert cell.num_trainable_params == 1

    cell.get_parameters()


def test_make_trainable_network(SimpleCell):
    """Test make_trainable."""
    cell = SimpleCell(4, 4)
    cell.insert(HH())

    net = jx.Network([cell, cell])
    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(0).loc(0.0),
        IonotropicSynapse(),
    )

    cell.branch(0).loc(0.0).set("length", 12.0)
    cell.branch(1).loc(1.0).set("HH_gNa", 0.2)

    cell.branch([0, 1]).make_trainable("radius", 1.0)
    cell.branch([0, 1]).make_trainable("length")
    cell.branch([0, 1]).make_trainable("axial_resistivity", [600.0, 700.0])
    cell.branch([0, 1]).make_trainable("HH_gNa")

    cell.get_parameters()
    net.IonotropicSynapse.set("IonotropicSynapse_gS", 0.1)
    assert cell.num_trainable_params == 8  # `set()` is ignored.


def test_diverse_synapse_types(SimpleNet):
    """Runs `.get_all_parameters()` and checks if the output is as expected."""
    net = SimpleNet(4, 1, 1)
    for pre_ind in [0, 1]:
        for post_ind, syn in zip([2, 3], [IonotropicSynapse(), TestSynapse()]):
            pre = net.cell(pre_ind).branch(0).loc(0.0)
            post = net.cell(post_ind).branch(0).loc(0.0)
            connect(pre, post, syn)

    net.IonotropicSynapse.make_trainable("IonotropicSynapse_gS")
    net.TestSynapse.edge([0, 1]).make_trainable("TestSynapse_gC")
    assert net.num_trainable_params == 3

    params = net.get_parameters()

    # Modify the trainable parameters.
    params[0]["IonotropicSynapse_gS"] = params[0]["IonotropicSynapse_gS"].at[:].set(2.2)
    params[1]["TestSynapse_gC"] = params[1]["TestSynapse_gC"].at[0].set(3.3)
    params[1]["TestSynapse_gC"] = params[1]["TestSynapse_gC"].at[1].set(4.4)
    net.to_jax()
    pstate = params_to_pstate(params, net.indices_set_by_trainables)
    all_parameters = net.get_all_parameters(pstate, voltage_solver="jaxley.thomas")

    assert np.all(all_parameters["radius"] == 1.0)
    assert np.all(all_parameters["length"] == 10.0)
    assert np.all(all_parameters["axial_resistivity"] == 5000.0)
    assert np.all(all_parameters["IonotropicSynapse_gS"][0] == 2.2)
    assert np.all(all_parameters["IonotropicSynapse_gS"][1] == 2.2)
    assert np.all(all_parameters["TestSynapse_gC"][0] == 3.3)
    assert np.all(all_parameters["TestSynapse_gC"][1] == 4.4)

    # Add another trainable parameter and test again.
    net.IonotropicSynapse.edge(1).make_trainable("IonotropicSynapse_gS")
    assert net.num_trainable_params == 4

    params = net.get_parameters()

    # Modify the trainable parameters.
    params[2]["IonotropicSynapse_gS"] = params[2]["IonotropicSynapse_gS"].at[:].set(5.5)
    net.to_jax()
    pstate = params_to_pstate(params, net.indices_set_by_trainables)
    all_parameters = net.get_all_parameters(pstate, voltage_solver="jaxley.thomas")
    assert np.all(all_parameters["IonotropicSynapse_gS"][0] == 2.2)
    assert np.all(all_parameters["IonotropicSynapse_gS"][1] == 5.5)


def test_make_all_trainable_corresponds_to_set(SimpleNet):
    # Scenario 1.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.insert(HH())
    params1 = get_params_all_trainable(net1)
    net2.insert(HH())
    params2 = get_params_set(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)

    # Scenario 2.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(1).insert(HH())
    params1 = get_params_all_trainable(net1)
    net2.cell(1).insert(HH())
    params2 = get_params_set(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)

    # Scenario 3.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(1).branch(0).insert(HH())
    params1 = get_params_all_trainable(net1)
    net2.cell(1).branch(0).insert(HH())
    params2 = get_params_set(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)

    # Scenario 4.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(1).branch(0).loc(0.4).insert(HH())
    params1 = get_params_all_trainable(net1)
    net2.cell(1).branch(0).loc(0.4).insert(HH())
    params2 = get_params_set(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)


def test_make_subset_trainable_corresponds_to_set(SimpleNet):
    # Scenario 1.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.insert(HH())
    params1 = get_params_subset_trainable(net1)
    net2.insert(HH())
    params2 = get_params_set_subset(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)

    # Scenario 2.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(0).insert(HH())
    params1 = get_params_subset_trainable(net1)
    net2.cell(0).insert(HH())
    params2 = get_params_set_subset(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)

    # Scenario 3.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(0).branch(1).insert(HH())
    params1 = get_params_subset_trainable(net1)
    net2.cell(0).branch(1).insert(HH())
    params2 = get_params_set_subset(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)

    # Scenario 4.
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(0).branch(1).loc(0.4).insert(HH())
    params1 = get_params_subset_trainable(net1)
    net2.cell(0).branch(1).loc(0.4).insert(HH())
    params2 = get_params_set_subset(net2)
    assert np.array_equal(params1["HH_gNa"], params2["HH_gNa"], equal_nan=True)


def test_copy_node_property_to_edges(SimpleNet):
    """Test synaptic parameter sharing via `.copy_node_property_to_edges()`.

    This test does not explicitly use `make_trainable`, but
    `copy_node_property_to_edges` is an important ingredient to parameter sharing.
    """
    net = SimpleNet(6, 2, 2)
    net.insert(HH())
    net.cell(1).set("HH_gNa", 1.0)
    net.cell(0).set("radius", 0.2)
    net.cell(1).branch(0).comp(0).set("capacitance", 0.3)
    fully_connect(net.cell("all"), net.cell("all"), IonotropicSynapse())

    net.copy_node_property_to_edges("HH_gNa", "pre")
    # Run it another time to ensure that it can be run twice.
    net.copy_node_property_to_edges("HH_gNa", "pre")
    assert "pre_HH_gNa" in net.edges.columns
    assert "post_HH_gNa" not in net.edges.columns

    # Query the second cell. Each cell has four compartments.
    edges_gna_values = net.edges.query("pre_global_comp_index > 3")
    edges_gna_values = edges_gna_values.query("pre_global_comp_index <= 7")
    assert np.all(edges_gna_values["pre_HH_gNa"] == 1.0)

    # Query the other cells. The first cell has four compartments.
    edges_gna_values = net.edges.query("pre_global_comp_index <= 3")
    assert np.all(edges_gna_values["pre_HH_gNa"] == 0.12)
    edges_gna_values = net.edges.query("pre_global_comp_index > 7")
    assert np.all(edges_gna_values["pre_HH_gNa"] == 0.12)

    # Test whether multiple properties can be copied over.
    net.copy_node_property_to_edges(["radius", "length"])
    assert "pre_radius" in net.edges.columns
    assert "post_radius" in net.edges.columns
    assert "pre_length" in net.edges.columns
    assert "post_length" in net.edges.columns

    edges_gna_values = net.edges.query("pre_global_comp_index <= 3")
    assert np.all(edges_gna_values["pre_radius"] == 0.2)

    edges_gna_values = net.edges.query("pre_global_comp_index > 3")
    assert np.all(edges_gna_values["pre_radius"] == 1.0)

    # Test whether modifying an individual compartment also takes effect.
    net.copy_node_property_to_edges(["capacitance"])
    assert "pre_capacitance" in net.edges.columns
    assert "post_capacitance" in net.edges.columns

    edges_gna_values = net.edges.query("pre_global_comp_index == 4")
    assert np.all(edges_gna_values["pre_capacitance"] == 0.3)

    edges_gna_values = net.edges.query("post_global_comp_index == 4")
    assert np.all(edges_gna_values["post_capacitance"] == 0.3)

    edges_gna_values = net.edges.query("pre_global_comp_index != 4")
    assert np.all(edges_gna_values["pre_capacitance"] == 1.0)

    edges_gna_values = net.edges.query("post_global_comp_index != 4")
    assert np.all(edges_gna_values["post_capacitance"] == 1.0)


def get_params_subset_trainable(net):
    net.cell(0).branch(1).make_trainable("HH_gNa")
    params = net.get_parameters()
    params[0]["HH_gNa"] = params[0]["HH_gNa"].at[:].set(0.0)
    net.to_jax()
    pstate = params_to_pstate(params, net.indices_set_by_trainables)
    return net.get_all_parameters(pstate, voltage_solver="jaxley.thomas")


def get_params_set_subset(net):
    net.cell(0).branch(1).set("HH_gNa", 0.0)
    params = net.get_parameters()
    net.to_jax()
    pstate = params_to_pstate(params, net.indices_set_by_trainables)
    return net.get_all_parameters(pstate, voltage_solver="jaxley.thomas")


def get_params_all_trainable(net):
    net.cell("all").branch("all").loc("all").make_trainable("HH_gNa")
    params = net.get_parameters()
    params[0]["HH_gNa"] = params[0]["HH_gNa"].at[:].set(0.0)
    net.to_jax()
    pstate = params_to_pstate(params, net.indices_set_by_trainables)
    return net.get_all_parameters(pstate, voltage_solver="jaxley.thomas")


def get_params_set(net):
    net.set("HH_gNa", 0.0)
    params = net.get_parameters()
    net.to_jax()
    pstate = params_to_pstate(params, net.indices_set_by_trainables)
    return net.get_all_parameters(pstate, voltage_solver="jaxley.thomas")


def test_make_trainable_corresponds_to_set_pospischil(SimpleNet):
    """Test whether shared parameters are also set correctly."""
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(0).insert(Na())
    net1.insert(K())
    net1.cell("all").branch("all").loc("all").make_trainable("vt")
    params1 = net1.get_parameters()
    params1[0]["vt"] = params1[0]["vt"].at[:].set(0.05)
    net1.to_jax()
    pstate1 = params_to_pstate(params1, net1.indices_set_by_trainables)
    all_params1 = net1.get_all_parameters(pstate1, voltage_solver="jaxley.thomas")

    net2.cell(0).insert(Na())
    net2.insert(K())
    net2.cell("all").branch("all").loc("all").make_trainable("vt")
    params2 = net2.get_parameters()
    params2[0]["vt"] = params2[0]["vt"].at[:].set(0.05)
    net2.to_jax()
    pstate2 = params_to_pstate(params2, net2.indices_set_by_trainables)
    all_params2 = net2.get_all_parameters(pstate2, voltage_solver="jaxley.thomas")
    assert np.array_equal(all_params1["vt"], all_params2["vt"], equal_nan=True)
    assert np.array_equal(all_params1["Na_gNa"], all_params2["Na_gNa"], equal_nan=True)
    assert np.array_equal(all_params1["K_gK"], all_params2["K_gK"], equal_nan=True)

    # Perform test on simulation.
    net1.cell(1).branch(1).loc(0.0).record()
    net2.cell(1).branch(1).loc(0.0).record()

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    net1.cell(0).branch(1).loc(0.0).stimulate(current)
    net2.cell(0).branch(1).loc(0.0).stimulate(current)
    voltages1 = jx.integrate(net1, params=params1)
    voltages2 = jx.integrate(net2, params=params2)
    assert np.max(np.abs(voltages1 - voltages2)) < 1e-8


def test_group_trainable_corresponds_to_set():
    """Use `GroupView` and make it trainable; test if it gives the same as `set`."""

    def build_net():
        comp = jx.Compartment()
        branch = jx.Branch(comp, ncomp=4)
        cell = jx.Cell(branch, parents=[-1, 0, 0, 1, 1])
        net = jx.Network([cell for _ in range(4)])
        net.cell(0).add_to_group("test")
        net.cell(1).branch(2).add_to_group("test")
        return net

    net1 = build_net()

    net1.test.make_trainable("radius")
    params = net1.get_parameters()
    params[0]["radius"] = params[0]["radius"].at[:].set(2.5)
    net1.to_jax()
    pstate = params_to_pstate(params, net1.indices_set_by_trainables)
    all_parameters1 = net1.get_all_parameters(pstate, voltage_solver="jaxley.thomas")

    net2 = build_net()
    net2.test.set("radius", 2.5)
    params = net2.get_parameters()
    net2.to_jax()
    pstate = params_to_pstate(params, net2.indices_set_by_trainables)
    all_parameters2 = net2.get_all_parameters(pstate, voltage_solver="jaxley.thomas")

    assert np.allclose(all_parameters1["radius"], all_parameters2["radius"])


def test_data_set_vs_make_trainable_pospischil(SimpleNet):
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    net1.cell(0).insert(Na())
    net1.insert(K())
    net1.make_trainable("vt")
    params1 = net1.get_parameters()
    params1[0]["vt"] = params1[0]["vt"].at[:].set(0.05)
    net1.to_jax()
    pstate1 = params_to_pstate(params1, net1.indices_set_by_trainables)
    all_params1 = net1.get_all_parameters(pstate1, voltage_solver="jaxley.thomas")

    net2.cell(0).insert(Na())
    net2.insert(K())
    val = params1[0]["vt"]
    pstate = net2.cell("all").branch("all").loc("all").data_set("vt", val.item(), None)
    net2.to_jax()
    all_params2 = net2.get_all_parameters(pstate, voltage_solver="jaxley.thomas")
    assert np.array_equal(all_params1["vt"], all_params2["vt"], equal_nan=True)
    assert np.array_equal(all_params1["Na_gNa"], all_params2["Na_gNa"], equal_nan=True)
    assert np.array_equal(all_params1["K_gK"], all_params2["K_gK"], equal_nan=True)

    # Perform test on simulation.
    net1.cell(1).branch(1).loc(0.0).record()
    net2.cell(1).branch(1).loc(0.0).record()

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    net1.cell(0).branch(1).loc(0.0).stimulate(current)
    net2.cell(0).branch(1).loc(0.0).stimulate(current)
    voltages1 = jx.integrate(net1, params=params1)
    voltages2 = jx.integrate(net2, param_state=pstate)
    assert np.max(np.abs(voltages1 - voltages2)) < 1e-8


def test_data_set_vs_make_trainable_network(SimpleNet):
    net1 = SimpleNet(2, 4, 1)
    net2 = SimpleNet(2, 4, 1)
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    for net in [net1, net2]:
        net.insert(HH())
        pre = net.cell(0).branch(0).loc(0.0)
        post = net.cell(0).branch(1).loc(0.9)
        connect(pre, post, IonotropicSynapse())
        pre = net.cell(0).branch(0).loc(0.1)
        post = net.cell(1).branch(1).loc(0.4)
        connect(pre, post, TestSynapse())
        pre = net.cell(1).branch(0).loc(0.3)
        post = net.cell(0).branch(1).loc(0.0)
        connect(pre, post, IonotropicSynapse())

        net.cell(0).branch(1).loc(1.0).record()
        net.cell(1).branch(0).loc(1.0).record()

        net.cell(1).branch(0).loc(0.4).stimulate(current)
        net.cell(0).branch(0).loc(0.6).stimulate(current)

    net1.make_trainable("radius", 0.9)
    net1.make_trainable("length", 0.99)
    net1.IonotropicSynapse.edge("all").make_trainable("IonotropicSynapse_gS", 0.15)
    net1.IonotropicSynapse.edge(1).make_trainable("IonotropicSynapse_e_syn", 0.2)
    net1.TestSynapse.edge(0).make_trainable("TestSynapse_gC", 0.3)
    params1 = net1.get_parameters()

    pstate = None
    pstate = net2.data_set("radius", 0.9, pstate)
    pstate = net2.data_set("length", 0.99, pstate)
    pstate = net2.IonotropicSynapse.edge("all").data_set(
        "IonotropicSynapse_gS", 0.15, pstate
    )
    pstate = net2.IonotropicSynapse.edge(1).data_set(
        "IonotropicSynapse_e_syn", 0.2, pstate
    )
    pstate = net2.TestSynapse.edge(0).data_set("TestSynapse_gC", 0.3, pstate)

    voltages1 = jx.integrate(net1, params=params1)
    voltages2 = jx.integrate(net2, param_state=pstate)
    assert np.max(np.abs(voltages1 - voltages2)) < 1e-8


def test_make_states_trainable_api(SimpleNet):
    net = SimpleNet(2, 2, 4)
    net.insert(HH())
    net.cell(0).branch(0).comp(0).record()

    net.cell("all").make_trainable("v")
    net.make_trainable("HH_h")
    net.cell(0).branch("all").make_trainable("HH_m")
    net.cell(0).branch("all").comp("all").make_trainable("HH_n")

    def simulate(params):
        return jx.integrate(net, params=params, t_max=10.0)

    parameters = net.get_parameters()
    v = simulate(parameters)
    assert np.invert(np.any(np.isnan(v))), "Found NaN in voltage."


def test_write_trainables(SimpleNet):
    """Test whether `write_trainables()` gives the same result as using the trainables."""
    net = SimpleNet(2, 2, 4)
    connect(
        net.cell(0).branch(0).loc(0.9),
        net.cell(1).branch(1).loc(0.1),
        IonotropicSynapse(),
    )
    connect(
        net.cell(1).branch(0).loc(0.1),
        net.cell(0).branch(1).loc(0.3),
        TestSynapse(),
    )
    connect(
        net.cell(0).branch(0).loc(0.3),
        net.cell(0).branch(1).loc(0.6),
        TestSynapse(),
    )
    connect(
        net.cell(1).branch(0).loc(0.6),
        net.cell(1).branch(1).loc(0.9),
        IonotropicSynapse(),
    )
    net.insert(HH())
    net.cell(0).branch(0).comp(0).record()
    net.cell(1).branch(0).comp(0).record()
    net.cell(0).branch(0).comp(0).stimulate(
        jx.step_current(i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0)
    )

    net.make_trainable("radius")
    net.cell(0).make_trainable("length")
    net.cell("all").make_trainable("axial_resistivity")
    net.cell("all").branch("all").make_trainable("HH_gNa")
    net.cell("all").branch("all").make_trainable("HH_m")
    net.make_trainable("IonotropicSynapse_gS")
    net.make_trainable("IonotropicSynapse_s")
    net.select(edges="all").make_trainable("TestSynapse_gC")
    net.select(edges="all").make_trainable("TestSynapse_c")
    net.cell(0).branch(0).comp(0).make_trainable("radius")

    params = net.get_parameters()

    # Now, we manually modify the parameters.
    for p in params:
        for key in p:
            p[key] = p[key].at[:].set(np.random.rand())

    # Test whether voltages match.
    v1 = jx.integrate(net, params=params)

    previous_nodes = copy(net.nodes)
    previous_edges = copy(net.edges)
    net.write_trainables(params)
    v2 = jx.integrate(net)
    assert np.max(np.abs(v1 - v2)) < 1e-8

    # Test whether nodes and edges actually changed.
    assert not net.nodes.equals(previous_nodes)
    assert not net.edges.equals(previous_edges)

    # Test whether `View` raises with `write_trainables()`.
    with pytest.raises(AssertionError):
        net.cell(0).write_trainables(params)

    # Test whether synapse view raises an error.
    with pytest.raises(AssertionError):
        net.select(edges=[0, 2, 3]).write_trainables(params)


def test_param_sharing_w_different_group_sizes():
    # test if make_trainable corresponds to set
    branch1 = jx.Branch(ncomp=6)
    branch1.nodes["controlled_by_param"] = np.array([0, 0, 0, 1, 1, 2])
    branch1.make_trainable("radius")
    assert branch1.num_trainable_params == 3

    # make trainable
    params = branch1.get_parameters()
    params[0]["radius"] = params[0]["radius"].at[:].set([2, 3, 4])
    branch1.to_jax()
    pstate = params_to_pstate(params, branch1.indices_set_by_trainables)
    params1 = branch1.get_all_parameters(pstate, voltage_solver="jaxley.thomas")

    # set
    branch2 = jx.Branch(ncomp=6)
    branch2.set("radius", np.array([2, 2, 2, 3, 3, 4]))
    params = branch2.get_parameters()
    branch2.to_jax()
    pstate = params_to_pstate(params, branch2.indices_set_by_trainables)
    params2 = branch2.get_all_parameters(pstate, voltage_solver="jaxley.thomas")

    assert np.array_equal(params1["radius"], params2["radius"], equal_nan=True)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH


@pytest.mark.parametrize(
    "property", ["radius", "capacitance", "length", "axial_resistivity"]
)
def test_raise_for_heterogenous_modules(SimpleBranch, property):
    branch0 = SimpleBranch(4)
    branch1 = SimpleBranch(4)
    branch1.comp(1).set(property, 1.5)
    cell = jx.Cell([branch0, branch1], parents=[-1, 0])
    with pytest.raises(ValueError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_heterogenous_channel_existance(SimpleBranch):
    branch0 = SimpleBranch(4)
    branch1 = SimpleBranch(4)
    branch1.comp(2).insert(HH())
    cell = jx.Cell([branch0, branch1], parents=[-1, 0])
    with pytest.raises(ValueError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_heterogenous_channel_properties(SimpleBranch):
    branch0 = SimpleBranch(4)
    branch1 = SimpleBranch(4)
    branch1.insert(HH())
    branch1.comp(3).set("HH_gNa", 0.5)
    cell = jx.Cell([branch0, branch1], parents=[-1, 0])
    with pytest.raises(ValueError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_entire_cells(SimpleCell):
    cell = SimpleCell(3, 4)
    with pytest.raises(AssertionError):
        cell.set_ncomp(2)


def test_raise_for_networks(SimpleCell):
    cell1 = SimpleCell(3, 4)
    cell2 = SimpleCell(3, 4)
    net = jx.Network([cell1, cell2])
    with pytest.raises(AssertionError):
        net.cell(0).branch(1).set_ncomp(2)


def test_raise_for_recording(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.branch(0).comp(0).record()
    with pytest.raises(AssertionError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_stimulus(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.branch(0).comp(0).stimulate(0.4 * jnp.ones(100))
    with pytest.raises(AssertionError):
        cell.branch(1).set_ncomp(2)


@pytest.mark.parametrize("new_ncomp", [1, 2, 4, 5, 8])
def test_simulation_accuracy_api_equivalence_init_vs_setncomp_branch(
    SimpleBranch, new_ncomp
):
    """Test whether a module built from scratch matches module built with `set_ncomp()`.

    This makes one branch, whose `ncomp` is not modified, heterogenous.
    """
    branch1 = SimpleBranch(new_ncomp)

    # The second branch is originally instantiated to have 4 ncomp, but is later
    # modified to have `new_ncomp` compartments.
    branch2 = SimpleBranch(4)
    branch2.comp("all").set("length", 10.0)
    total_branch_len = 4 * 10.0

    # Make the total branch length 40 um.
    branch1.comp("all").set("length", total_branch_len / new_ncomp)

    # Adapt ncomp.
    branch2.set_ncomp(new_ncomp)

    for branch in [branch1, branch2]:
        branch.comp(0).stimulate(0.4 * jnp.ones(100))
        branch.comp(new_ncomp - 1).record()

    v1 = jx.integrate(branch1)
    v2 = jx.integrate(branch2)
    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-8, f"Too large voltage deviation, {max_error} > 1e-8"


@pytest.mark.parametrize("new_ncomp", [1, 2, 4, 5, 8])
def test_simulation_accuracy_api_equivalence_init_vs_setncomp_cell(
    SimpleBranch, new_ncomp
):
    """Test whether a module built from scratch matches module built with `set_ncomp()`."""
    branch1 = SimpleBranch(new_ncomp)

    # The second branch is originally instantiated to have 4 ncomp, but is later
    # modified to have `new_ncomp` compartments.
    branch2 = SimpleBranch(4)
    branch2.comp("all").set("length", 10.0)
    total_branch_len = 4 * 10.0

    # Make the total branch length 20 um.
    branch1.comp("all").set("length", total_branch_len / new_ncomp)
    cell1 = jx.Cell(branch1, parents=[-1, 0])
    cell2 = jx.Cell(branch2, parents=[-1, 0])

    # Adapt ncomp.
    for b in range(2):
        cell2.branch(b).set_ncomp(new_ncomp)

    for cell in [cell1, cell2]:
        cell.branch(0).comp(0).stimulate(0.4 * jnp.ones(100))
        cell.branch(1).comp(new_ncomp - 1).record()

    v1 = jx.integrate(cell1)
    v2 = jx.integrate(cell2)
    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-8, f"Too large voltage deviation, {max_error} > 1e-8"


@pytest.mark.parametrize("new_ncomp", [1, 2, 4, 5, 8])
@pytest.mark.parametrize("file", ["morph_250.swc"])
def test_api_equivalence_swc_lengths_and_radiuses(SimpleMorphCell, new_ncomp, file):
    """Test if the radiuses and lenghts of an SWC morph are reconstructed correctly."""
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    cell1 = SimpleMorphCell(fname, ncomp=new_ncomp)
    cell2 = SimpleMorphCell(fname, ncomp=1)

    for b in range(cell2.total_nbranches):
        cell2.branch(b).set_ncomp(new_ncomp)

    for property_name in ["radius", "length"]:
        cell1_vals = cell1.nodes[property_name].to_numpy()
        cell2_vals = cell2.nodes[property_name].to_numpy()
        assert np.allclose(
            cell1_vals, cell2_vals
        ), f"Too large difference in {property_name}"


@pytest.mark.parametrize("new_ncomp", [1, 2, 4, 5, 8])
@pytest.mark.parametrize("file", ["morph_250.swc"])
def test_simulation_accuracy_swc_init_vs_set_ncomp(SimpleMorphCell, new_ncomp, file):
    """Test whether an SWC initially built with 4 ncomp works after `set_ncomp()`."""
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    cell1 = SimpleMorphCell(fname, ncomp=new_ncomp)
    cell2 = SimpleMorphCell(fname, ncomp=1)

    for b in range(cell2.total_nbranches):
        cell2.branch(b).set_ncomp(new_ncomp)

    for cell in [cell1, cell2]:
        cell.branch(0).comp(0).stimulate(0.4 * jnp.ones(100))
        cell.branch(0).comp(new_ncomp - 1).record()
        cell.branch(3).comp(0).record()
        cell.branch(5).comp(new_ncomp - 1).record()

    v1 = jx.integrate(cell1, voltage_solver="jax.sparse")
    v2 = jx.integrate(cell2, voltage_solver="jax.sparse")
    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-8, f"Too large voltage deviation, {max_error} > 1e-8"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Callable, Dict, List, Optional, Tuple, Union
from warnings import warn

import jax.numpy as jnp
import numpy as np
import pandas as pd

from jaxley.modules.base import Module
from jaxley.modules.compartment import Compartment
from jaxley.utils.cell_utils import compute_children_and_parents
from jaxley.utils.misc_utils import cumsum_leading_zero, deprecated_kwargs
from jaxley.utils.solver_utils import JaxleySolveIndexer, comp_edges_to_indices


class Branch(Module):
    """Branch class.

    This class defines a single branch that can be simulated by itself or
    connected to build a cell. A branch is linear segment of several compartments
    and can be connected to no, one or more other branches at each end to build more
    intricate cell morphologies.
    """

    branch_params: Dict = {}
    branch_states: Dict = {}

    def __init__(
        self,
        compartments: Optional[Union[Compartment, List[Compartment]]] = None,
        ncomp: Optional[int] = None,
    ):
        """
        Args:
            compartments: A single compartment or a list of compartments that make up the
                branch.
            ncomp: Number of segments to divide the branch into. If `compartments` is an
                a single compartment, than the compartment is repeated `ncomp` times to
                create the branch.
        """
        super().__init__()
        assert (
            isinstance(compartments, (Compartment, List)) or compartments is None
        ), "Only Compartment or List[Compartment] is allowed."
        if isinstance(compartments, Compartment):
            assert (
                ncomp is not None
            ), "If `compartments` is not a list then you have to set `ncomp`."
        compartments = Compartment() if compartments is None else compartments
        ncomp = 1 if ncomp is None else ncomp

        if isinstance(compartments, Compartment):
            compartment_list = [compartments] * ncomp
        else:
            compartment_list = compartments

        self.ncomp = len(compartment_list)
        self.ncomp_per_branch = np.asarray([self.ncomp])
        self.total_nbranches = 1
        self.nbranches_per_cell = [1]
        self._cumsum_nbranches = jnp.asarray([0, 1])
        self.cumsum_ncomp = cumsum_leading_zero(self.ncomp_per_branch)

        # Indexing.
        self.nodes = pd.concat([c.nodes for c in compartment_list], ignore_index=True)
        self._append_params_and_states(self.branch_params, self.branch_states)
        self.nodes["global_comp_index"] = np.arange(self.ncomp).tolist()
        self.nodes["global_branch_index"] = [0] * self.ncomp
        self.nodes["global_cell_index"] = [0] * self.ncomp
        self._update_local_indices()
        self._init_view()

        # Channels.
        self._gather_channels_from_constituents(compartment_list)

        self.branch_edges = pd.DataFrame(
            dict(parent_branch_index=[], child_branch_index=[])
        )

        # For morphology indexing.
        self._par_inds, self._child_inds, self._child_belongs_to_branchpoint = (
            compute_children_and_parents(self.branch_edges)
        )
        self._internal_node_inds = jnp.arange(self.ncomp)

        self._initialize()

        # Coordinates.
        self.xyzr = [float("NaN") * np.zeros((2, 4))]

    def _init_morph_jaxley_spsolve(self):
        self._solve_indexer = JaxleySolveIndexer(
            cumsum_ncomp=self.cumsum_ncomp,
            ncomp_per_branch=self.ncomp_per_branch,
            branchpoint_group_inds=np.asarray([]).astype(int),
            remapped_node_indices=self._internal_node_inds,
            children_in_level=[],
            parents_in_level=[],
            root_inds=np.asarray([0]),
        )

    def _init_morph_jax_spsolve(self):
        """Initialize morphology for the jax sparse voltage solver.

        Explanation of `self._comp_eges['type']`:
        `type == 0`: compartment <--> compartment (within branch)
        `type == 1`: branchpoint --> parent-compartment
        `type == 2`: branchpoint --> child-compartment
        `type == 3`: parent-compartment --> branchpoint
        `type == 4`: child-compartment --> branchpoint
        """
        self._comp_edges = pd.DataFrame().from_dict(
            {
                "source": list(range(self.ncomp - 1)) + list(range(1, self.ncomp)),
                "sink": list(range(1, self.ncomp)) + list(range(self.ncomp - 1)),
            }
        )
        self._comp_edges["type"] = 0
        n_nodes, data_inds, indices, indptr = comp_edges_to_indices(self._comp_edges)
        self._n_nodes = n_nodes
        self._data_inds = data_inds
        self._indices_jax_spsolve = indices
        self._indptr_jax_spsolve = indptr

    def __len__(self) -> int:
        return self.ncomp



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, Optional, Tuple

import jax.numpy as jnp


class Synapse:
    """Base class for a synapse.

    As in NEURON, a `Synapse` is considered a point process, which means that its
    conductances are to be specified in `uS` and its currents are to be specified in
    `nA`.
    """

    _name = None
    synapse_params = None
    synapse_states = None

    def __init__(self, name: Optional[str] = None):
        self._name = name if name else self.__class__.__name__

    @property
    def name(self) -> Optional[str]:
        return self._name

    def change_name(self, new_name: str):
        """Change the synapse name.

        Args:
            new_name: The new name of the channel.

        Returns:
            Renamed channel, such that this function is chainable.
        """
        old_prefix = self._name + "_"
        new_prefix = new_name + "_"

        self._name = new_name
        self.synapse_params = {
            (
                new_prefix + key[len(old_prefix) :]
                if key.startswith(old_prefix)
                else key
            ): value
            for key, value in self.synapse_params.items()
        }

        self.synapse_states = {
            (
                new_prefix + key[len(old_prefix) :]
                if key.startswith(old_prefix)
                else key
            ): value
            for key, value in self.synapse_states.items()
        }
        return self

    def update_states(
        states: Dict[str, jnp.ndarray],
        delta_t: float,
        pre_voltage: jnp.ndarray,
        post_voltage: jnp.ndarray,
        params: Dict[str, jnp.ndarray],
    ) -> Dict[str, jnp.ndarray]:
        """ODE update step.

        Args:
            states: States of the synapse.
            delta_t: Time step in `ms`.
            pre_voltage: Voltage of the presynaptic compartment, shape `()`.
            post_voltage: Voltage of the postsynaptic compartment, shape `()`.
            params: Parameters of the synapse. Conductances in `uS`.

        Returns:
            Updated states."""
        raise NotImplementedError

    def compute_current(
        states: Dict[str, jnp.ndarray],
        pre_voltage: jnp.ndarray,
        post_voltage: jnp.ndarray,
        params: Dict[str, jnp.ndarray],
    ) -> jnp.ndarray:
        """Return current through one synapse in `nA`.

        Internally, we use `jax.vmap` to vectorize this function across many synapses.

        Args:
            states: States of the synapse.
            pre_voltage: Voltage of the presynaptic compartment, shape `()`.
            post_voltage: Voltage of the postsynaptic compartment, shape `()`.
            params: Parameters of the synapse. Conductances in `uS`.

        Returns:
            Current through the synapse in `nA`, shape `()`.
        """
        raise NotImplementedError



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'Jaxley'
copyright = '2024, Jaxleyverse team'
author = 'Jaxleyverse team'


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_math_dollar",
    "sphinx.ext.mathjax",
    "myst_nb",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "jax": ("https://jax.readthedocs.io/en/latest", None),
}

source_suffix = {
    '.rst': 'restructuredtext',
    '.myst': 'myst-nb',
    '.ipynb': 'myst-nb'
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Myst-NB
myst_enable_extensions = [
    "dollarmath",
    "amsmath",
    "deflist",
    "colon_fence",
]
nb_execution_timeout = 600
nb_execution_mode = "cache"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages. See the documentation for
# a list of builtin themes.
#
html_title = ""
html_logo = "logo.png"
html_theme = 'sphinx_book_theme'
html_theme_options = {
    'repository_url': 'https://github.com/jaxleyverse/jaxley',
    "use_repository_button": True,
    "use_download_button": False,
    'repository_branch': 'main',
    "path_to_docs": 'docs',
    'launch_buttons': {
        'colab_url': 'https://colab.research.google.com',
        'binderhub_url': 'https://mybinder.org'
    },
    "toc_title": "Navigation",
    "show_navbar_depth": 1,
    "show_toc_level": 3,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['custom.css']

autosummary_generate = True
autodoc_typehints = "description"
add_module_names = False
autodoc_member_order = "bysource"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import List

import jax.numpy as jnp
import numpy as np
from jax import vmap
from jax.experimental.sparse.linalg import spsolve as jax_spsolve
from tridiax.stone import stone_backsub_lower, stone_triang_upper
from tridiax.thomas import thomas_backsub_lower, thomas_triang_upper

from jaxley.utils.cell_utils import group_and_sum
from jaxley.utils.solver_utils import JaxleySolveIndexer


def step_voltage_explicit(
    voltages: jnp.ndarray,
    voltage_terms: jnp.ndarray,
    constant_terms: jnp.ndarray,
    axial_conductances: jnp.ndarray,
    internal_node_inds: jnp.ndarray,
    sinks: jnp.ndarray,
    sources: jnp.ndarray,
    types: jnp.ndarray,
    ncomp_per_branch: jnp.ndarray,
    par_inds: jnp.ndarray,
    child_inds: jnp.ndarray,
    nbranches: int,
    solver: str,
    idx: JaxleySolveIndexer,
    debug_states,
    delta_t: float,
) -> jnp.ndarray:
    """Solve one timestep of branched nerve equations with explicit (forward) Euler."""
    voltages = jnp.reshape(voltages, (nbranches, -1))
    voltage_terms = jnp.reshape(voltage_terms, (nbranches, -1))
    constant_terms = jnp.reshape(constant_terms, (nbranches, -1))

    update = _voltage_vectorfield(
        voltages,
        voltage_terms,
        constant_terms,
        types,
        sources,
        sinks,
        axial_conductances,
        par_inds,
        child_inds,
        nbranches,
        solver,
        delta_t,
        idx,
        debug_states,
    )
    new_voltates = voltages + delta_t * update
    return new_voltates.ravel(order="C")


def step_voltage_implicit_with_jaxley_spsolve(
    voltages: jnp.ndarray,
    voltage_terms: jnp.ndarray,
    constant_terms: jnp.ndarray,
    axial_conductances: jnp.ndarray,
    internal_node_inds: jnp.ndarray,
    sinks: jnp.ndarray,
    sources: jnp.ndarray,
    types: jnp.ndarray,
    ncomp_per_branch: jnp.ndarray,
    par_inds: jnp.ndarray,
    child_inds: jnp.ndarray,
    nbranches: int,
    solver: str,
    idx: JaxleySolveIndexer,
    debug_states,
    delta_t: float,
):
    """Solve one timestep of branched nerve equations with implicit (backward) Euler."""
    # Build diagonals.
    c2c = np.isin(types, [0, 1, 2])
    total_ncomp = idx.cumsum_ncomp[-1]
    diags = jnp.ones(total_ncomp)

    # if-case needed because `.at` does not allow empty inputs, but the input is
    # empty for compartments.
    if len(sinks[c2c]) > 0:
        diags = diags.at[idx.mask(sinks[c2c])].add(delta_t * axial_conductances[c2c])

    diags = diags.at[idx.mask(internal_node_inds)].add(delta_t * voltage_terms)

    # Build solves.
    solves = jnp.zeros(total_ncomp)
    solves = solves.at[idx.mask(internal_node_inds)].add(
        voltages + delta_t * constant_terms
    )

    # Build upper and lower within the branch.
    c2c = types == 0  # c2c = compartment-to-compartment.

    # Build uppers.
    uppers = jnp.zeros(total_ncomp)
    upper_inds = sources[c2c] > sinks[c2c]
    sinks_upper = sinks[c2c][upper_inds]
    if len(sinks_upper) > 0:
        uppers = uppers.at[idx.mask(sinks_upper)].add(
            -delta_t * axial_conductances[c2c][upper_inds]
        )

    # Build lowers.
    lowers = jnp.zeros(total_ncomp)
    lower_inds = sources[c2c] < sinks[c2c]
    sinks_lower = sinks[c2c][lower_inds]
    if len(sinks_lower) > 0:
        lowers = lowers.at[idx.mask(sinks_lower)].add(
            -delta_t * axial_conductances[c2c][lower_inds]
        )

    # Build branchpoint conductances.
    branchpoint_conds_parents = axial_conductances[types == 1]
    branchpoint_conds_children = axial_conductances[types == 2]
    branchpoint_weights_parents = axial_conductances[types == 3]
    branchpoint_weights_children = axial_conductances[types == 4]
    all_branchpoint_vals = jnp.concatenate(
        [branchpoint_weights_parents, branchpoint_weights_children]
    )
    # Find unique group identifiers
    num_branchpoints = len(branchpoint_conds_parents)
    branchpoint_diags = -group_and_sum(
        all_branchpoint_vals, idx.branchpoint_group_inds, num_branchpoints
    )
    branchpoint_solves = jnp.zeros((num_branchpoints,))

    branchpoint_conds_children = -delta_t * branchpoint_conds_children
    branchpoint_conds_parents = -delta_t * branchpoint_conds_parents

    # Here, I move all child and parent indices towards a branchpoint into a larger
    # vector. This is wasteful, but it makes indexing much easier. JIT compiling
    # makes the speed difference negligible.
    # Children.
    bp_conds_children = jnp.zeros(nbranches)
    bp_weights_children = jnp.zeros(nbranches)
    # Parents.
    bp_conds_parents = jnp.zeros(nbranches)
    bp_weights_parents = jnp.zeros(nbranches)

    # `.at[inds]` requires that `inds` is not empty, so we need an if-case here.
    # `len(inds) == 0` is the case for branches and compartments.
    if num_branchpoints > 0:
        bp_conds_children = bp_conds_children.at[child_inds].set(
            branchpoint_conds_children
        )
        bp_weights_children = bp_weights_children.at[child_inds].set(
            branchpoint_weights_children
        )
        bp_conds_parents = bp_conds_parents.at[par_inds].set(branchpoint_conds_parents)
        bp_weights_parents = bp_weights_parents.at[par_inds].set(
            branchpoint_weights_parents
        )

    # Triangulate the linear system of equations.
    (
        diags,
        lowers,
        solves,
        uppers,
        branchpoint_diags,
        branchpoint_solves,
        bp_weights_children,
        bp_conds_parents,
    ) = _triang_branched(
        lowers,
        diags,
        uppers,
        solves,
        bp_conds_children,
        bp_conds_parents,
        bp_weights_children,
        bp_weights_parents,
        branchpoint_diags,
        branchpoint_solves,
        solver,
        ncomp_per_branch,
        idx,
        debug_states,
    )

    # Backsubstitute the linear system of equations.
    (
        solves,
        lowers,
        diags,
        bp_weights_parents,
        branchpoint_solves,
        bp_conds_children,
    ) = _backsub_branched(
        lowers,
        diags,
        uppers,
        solves,
        bp_conds_children,
        bp_conds_parents,
        bp_weights_children,
        bp_weights_parents,
        branchpoint_diags,
        branchpoint_solves,
        solver,
        ncomp_per_branch,
        idx,
        debug_states,
    )
    return solves.ravel(order="C")[idx.mask(internal_node_inds)]


def step_voltage_implicit_with_jax_spsolve(
    voltages,
    voltage_terms,
    constant_terms,
    axial_conductances,
    internal_node_inds,
    sinks,
    data_inds,
    indices,
    indptr,
    n_nodes,
    delta_t,
):
    axial_conductances = delta_t * axial_conductances

    # Build diagonals.
    diagonal_values = jnp.zeros(n_nodes)

    # if-case needed because `.at` does not allow empty inputs, but the input is
    # empty for compartments.
    if len(sinks) > 0:
        diagonal_values = diagonal_values.at[sinks].add(axial_conductances)

    diagonal_values = diagonal_values.at[internal_node_inds].add(
        1.0 + delta_t * voltage_terms
    )

    # Concatenate diagonals and off-diagonals (which are just `-axial_conductances`).
    all_values = jnp.concatenate([diagonal_values, -axial_conductances])

    # Build solve.
    solves = jnp.zeros(n_nodes)
    solves = solves.at[internal_node_inds].add(voltages + delta_t * constant_terms)

    # Solve the voltage equations.
    solution = jax_spsolve(all_values[data_inds], indices, indptr, solves)[
        internal_node_inds
    ]
    return solution


def _voltage_vectorfield(
    voltages: jnp.ndarray,
    voltage_terms: jnp.ndarray,
    constant_terms: jnp.ndarray,
    types: jnp.ndarray,
    sources: jnp.ndarray,
    sinks: jnp.ndarray,
    axial_conductances: jnp.ndarray,
    par_inds: jnp.ndarray,
    child_inds: jnp.ndarray,
    nbranches: int,
    solver: str,
    delta_t: float,
    idx: JaxleySolveIndexer,
    debug_states,
) -> jnp.ndarray:
    """Evaluate the vectorfield of the nerve equation."""
    if np.sum(np.isin(types, [1, 2, 3, 4])) > 0:
        raise NotImplementedError(
            f"Forward Euler is not implemented for branched morphologies."
        )

    # Membrane current update.
    vecfield = -voltage_terms * voltages + constant_terms

    # Build upper and lower within the branch.
    c2c = types == 0  # c2c = compartment-to-compartment.

    # Build uppers.
    upper_inds = sources[c2c] > sinks[c2c]
    if len(upper_inds) > 0:
        uppers = axial_conductances[c2c][upper_inds]
    else:
        uppers = jnp.asarray([])

    # Build lowers.
    lower_inds = sources[c2c] < sinks[c2c]
    if len(lower_inds) > 0:
        lowers = axial_conductances[c2c][lower_inds]
    else:
        lowers = jnp.asarray([])

    # For networks consisting of branches.
    uppers = jnp.reshape(uppers, (nbranches, -1))
    lowers = jnp.reshape(lowers, (nbranches, -1))

    # Current through segments within the same branch.
    vecfield = vecfield.at[:, :-1].add((voltages[:, 1:] - voltages[:, :-1]) * uppers)
    vecfield = vecfield.at[:, 1:].add((voltages[:, :-1] - voltages[:, 1:]) * lowers)

    return vecfield


def _triang_branched(
    lowers,
    diags,
    uppers,
    solves,
    branchpoint_conds_children,
    branchpoint_conds_parents,
    branchpoint_weights_children,
    branchpoint_weights_parents,
    branchpoint_diags,
    branchpoint_solves,
    tridiag_solver,
    ncomp_per_branch,
    idx,
    debug_states,
):
    """Triangulation."""
    for cil, pil in zip(
        reversed(idx.children_in_level), reversed(idx.parents_in_level)
    ):
        diags, lowers, solves, uppers = _triang_level(
            cil[:, 0],
            lowers,
            diags,
            uppers,
            solves,
            tridiag_solver,
            idx,
        )
        (
            branchpoint_diags,
            branchpoint_solves,
            branchpoint_weights_children,
        ) = _eliminate_children_lower(
            cil,
            diags,
            solves,
            branchpoint_conds_children,
            branchpoint_weights_children,
            branchpoint_diags,
            branchpoint_solves,
            idx,
        )
        diags, solves, branchpoint_conds_parents = _eliminate_parents_upper(
            pil,
            diags,
            solves,
            branchpoint_conds_parents,
            branchpoint_weights_parents,
            branchpoint_diags,
            branchpoint_solves,
            ncomp_per_branch,
            idx,
        )
    # At last level, we do not want to eliminate anymore.
    diags, lowers, solves, uppers = _triang_level(
        idx.root_inds, lowers, diags, uppers, solves, tridiag_solver, idx
    )
    return (
        diags,
        lowers,
        solves,
        uppers,
        branchpoint_diags,
        branchpoint_solves,
        branchpoint_weights_children,
        branchpoint_conds_parents,
    )


def _backsub_branched(
    lowers,
    diags,
    uppers,
    solves,
    branchpoint_conds_children,
    branchpoint_conds_parents,
    branchpoint_weights_children,
    branchpoint_weights_parents,
    branchpoint_diags,
    branchpoint_solves,
    tridiag_solver,
    ncomp_per_branch,
    idx,
    debug_states,
):
    """
    Backsubstitution.
    """
    # At first level, we do not want to eliminate.
    solves, lowers, diags = _backsub_level(
        idx.root_inds,
        diags,
        lowers,
        solves,
        tridiag_solver,
        idx,
    )
    counter = 0
    for cil, pil in zip(idx.children_in_level, idx.parents_in_level):
        branchpoint_weights_parents, branchpoint_solves = _eliminate_parents_lower(
            pil,
            diags,
            solves,
            branchpoint_weights_parents,
            branchpoint_solves,
            ncomp_per_branch,
            idx,
        )
        branchpoint_conds_children, solves = _eliminate_children_upper(
            cil,
            solves,
            branchpoint_conds_children,
            branchpoint_diags,
            branchpoint_solves,
            idx,
        )
        solves, lowers, diags = _backsub_level(
            cil[:, 0], diags, lowers, solves, tridiag_solver, idx
        )
        counter += 1
    return (
        solves,
        lowers,
        diags,
        branchpoint_weights_parents,
        branchpoint_solves,
        branchpoint_conds_children,
    )


def _triang_level(cil, lowers, diags, uppers, solves, tridiag_solver, idx):
    if tridiag_solver == "jaxley.stone":
        triang_fn = stone_triang_upper
    elif tridiag_solver == "jaxley.thomas":
        triang_fn = thomas_triang_upper
    else:
        raise NameError
    new_diags, new_lowers, new_solves = vmap(triang_fn, in_axes=(0, 0, 0, 0))(
        lowers[idx.lower(cil)],
        diags[idx.branch(cil)],
        uppers[idx.upper(cil)],
        solves[idx.branch(cil)],
    )
    diags = diags.at[idx.branch(cil)].set(new_diags)
    lowers = lowers.at[idx.lower(cil)].set(new_lowers)
    solves = solves.at[idx.branch(cil)].set(new_solves)
    uppers = uppers.at[idx.upper(cil)].set(0.0)

    return diags, lowers, solves, uppers


def _backsub_level(
    cil: jnp.ndarray,
    diags: jnp.ndarray,
    lowers: jnp.ndarray,
    solves: jnp.ndarray,
    tridiag_solver: str,
    idx,
) -> jnp.ndarray:
    bil = cil
    if tridiag_solver == "jaxley.stone":
        backsub_fn = stone_backsub_lower
    elif tridiag_solver == "jaxley.thomas":
        backsub_fn = thomas_backsub_lower
    else:
        raise NameError
    solves = solves.at[idx.branch(bil)].set(
        vmap(backsub_fn, in_axes=(0, 0, 0))(
            solves[idx.branch(bil)], lowers[idx.lower(bil)], diags[idx.branch(bil)]
        )
    )
    lowers = lowers.at[idx.lower(bil)].set(0.0)
    diags = diags.at[idx.branch(bil)].set(1.0)
    return solves, lowers, diags


def _eliminate_children_lower(
    cil,
    diags,
    solves,
    branchpoint_conds_children,
    branchpoint_weights_children,
    branchpoint_diags,
    branchpoint_solves,
    idx,
):
    bil = cil[:, 0]
    bpil = cil[:, 1]
    new_diag, new_solve = vmap(_eliminate_single_child_lower, in_axes=(0, 0, 0, 0))(
        diags[idx.first(bil)],
        solves[idx.first(bil)],
        branchpoint_conds_children[bil],
        branchpoint_weights_children[bil],
    )
    branchpoint_diags = branchpoint_diags.at[bpil].add(new_diag)
    branchpoint_solves = branchpoint_solves.at[bpil].add(new_solve)
    branchpoint_weights_children = branchpoint_weights_children.at[bil].set(0.0)
    return branchpoint_diags, branchpoint_solves, branchpoint_weights_children


def _eliminate_single_child_lower(
    diags,
    solves,
    branchpoint_conds_children,
    branchpoint_weights_children,
):
    multiplying_factor = -branchpoint_weights_children / diags

    update_diag = multiplying_factor * branchpoint_conds_children
    update_solve = multiplying_factor * solves
    return update_diag, update_solve


def _eliminate_parents_upper(
    pil,
    diags,
    solves,
    branchpoint_conds_parents,
    branchpoint_weights_parents,
    branchpoint_diags,
    branchpoint_solves,
    ncomp_per_branch: jnp.ndarray,
    idx,
):
    bil = pil[:, 0]
    bpil = pil[:, 1]
    new_diag, new_solve = vmap(_eliminate_single_parent_upper, in_axes=(0, 0, 0, 0))(
        branchpoint_conds_parents[bil],
        branchpoint_weights_parents[bil],
        branchpoint_diags[bpil],
        branchpoint_solves[bpil],
    )

    # Update the diagonal elements and `b` in `Ax=b` (called `solves`).
    diags = diags.at[idx.last(bil)].add(new_diag)
    solves = solves.at[idx.last(bil)].add(new_solve)
    branchpoint_conds_parents = branchpoint_conds_parents.at[bil].set(0.0)

    return diags, solves, branchpoint_conds_parents


def _eliminate_single_parent_upper(
    branchpoint_conds_parents,
    branchpoint_weights_parents,
    branchpoint_diags,
    branchpoint_solves,
):
    multiplying_factor = branchpoint_conds_parents / branchpoint_diags

    update_diag = -multiplying_factor * branchpoint_weights_parents
    update_solve = -multiplying_factor * branchpoint_solves
    return update_diag, update_solve


def _eliminate_parents_lower(
    pil,
    diags,
    solves,
    branchpoint_weights_parents,
    branchpoint_solves,
    ncomp_per_branch: jnp.ndarray,
    idx,
):
    bil = pil[:, 0]
    bpil = pil[:, 1]
    branchpoint_solves = branchpoint_solves.at[bpil].add(
        -solves[idx.last(bil)] * branchpoint_weights_parents[bil] / diags[idx.last(bil)]
    )
    branchpoint_weights_parents = branchpoint_weights_parents.at[bil].set(0.0)
    return branchpoint_weights_parents, branchpoint_solves


def _eliminate_children_upper(
    cil,
    solves,
    branchpoint_conds_children,
    branchpoint_diags,
    branchpoint_solves,
    idx,
):
    bil = cil[:, 0]
    bpil = cil[:, 1]
    solves = solves.at[idx.first(bil)].add(
        -branchpoint_solves[bpil]
        * branchpoint_conds_children[bil]
        / branchpoint_diags[bpil]
    )
    branchpoint_conds_children = branchpoint_conds_children.at[bil].set(0.0)
    return branchpoint_conds_children, solves



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from math import prod
from typing import Callable, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import pandas as pd

from jaxley.modules import Module
from jaxley.utils.cell_utils import params_to_pstate
from jaxley.utils.jax_utils import nested_checkpoint_scan


def build_init_and_step_fn(
    module: Module,
    voltage_solver: str = "jaxley.stone",
    solver: str = "bwd_euler",
) -> Tuple[Callable, Callable]:
    """This function returns the `init_fn` and `step_fn` which initialize the
    parameters and states of the neuron model and then step through the model

    Args:
        module (Module): A `Module` object that e.g. a cell.
        voltage_solver (str, optional): Voltage solver used in step. Defaults to "jaxley.stone".
        solver (str, optional): ODE solver. Defaults to "bwd_euler".

    Returns:
        init_fn, step_fn: Functions that initialize the state and parameters, and perform
            a single integration step, respectively.
    """
    # Initialize the external inputs and their indices.
    external_inds = module.external_inds.copy()

    def init_fn(
        params: List[Dict[str, jnp.ndarray]],
        all_states: Optional[Dict] = None,
        param_state: Optional[List[Dict]] = None,
        delta_t: float = 0.025,
    ) -> Tuple[Dict, Dict]:
        """Initializes the parameters and states of the neuron model.

        Args:
            params (List[Dict[str, jnp.ndarray]]): List of trainable parameters.
            all_states (Optional[Dict], optional): State if alread initialized. Defaults to None.
            param_state (Optional[List[Dict]], optional): Parameters returned by `data_set`.. Defaults to None.
            delta_t (float, optional): Step size. Defaults to 0.025.

        Returns:
            Tuple[Dict, Dict]: All states and parameters.
        """
        # Make the `trainable_params` of the same shape as the `param_state`, such that
        # they can be processed together by `get_all_parameters`.
        pstate = params_to_pstate(params, module.indices_set_by_trainables)
        if param_state is not None:
            pstate += param_state

        all_params = module.get_all_parameters(pstate, voltage_solver=voltage_solver)
        all_states = (
            module.get_all_states(pstate, all_params, delta_t)
            if all_states is None
            else all_states
        )
        return all_states, all_params

    def step_fn(
        all_states: Dict,
        all_params: Dict,
        externals: Dict,
        external_inds: Dict = external_inds,
        delta_t: float = 0.025,
    ) -> Dict:
        """Performs a single integration step with step size delta_t.

        Args:
            all_states (Dict): Current state of the neuron model.
            all_params (Dict): Current parameters of the neuron model.
            externals (Dict): External inputs.
            external_inds (Dict, optional): External indices. Defaults to `module.external_inds`.
            delta_t (float, optional): Time step. Defaults to 0.025.

        Returns:
            Dict: Updated states.
        """
        state = all_states
        state = module.step(
            state,
            delta_t,
            external_inds,
            externals,
            params=all_params,
            solver=solver,
            voltage_solver=voltage_solver,
        )
        return state

    return init_fn, step_fn


def add_stimuli(
    externals: Dict,
    external_inds: Dict,
    data_stimuli: Optional[Tuple[jnp.ndarray, pd.DataFrame]] = None,
) -> Tuple[Dict, Dict]:
    """Extends the external inputs with the stimuli.

    Args:
        externals (Dict): Current external inputs.
        external_inds (Dict): Current external indices.
        data_stimuli (Optional[Tuple[jnp.ndarray, pd.DataFrame]], optional): Additional data stimuli. Defaults to None.

    Returns:
        Tuple[Dict, Dict]: Updated external inputs and indices.
    """
    # If stimulus is inserted, add it to the external inputs.
    if "i" in externals.keys() or data_stimuli is not None:
        if "i" in externals.keys():
            if data_stimuli is not None:
                externals["i"] = jnp.concatenate([externals["i"], data_stimuli[1]])
                external_inds["i"] = jnp.concatenate(
                    [external_inds["i"], data_stimuli[2].index.to_numpy()]
                )
        else:
            externals["i"] = data_stimuli[1]
            external_inds["i"] = data_stimuli[2].index.to_numpy()

    return externals, external_inds


def add_clamps(
    externals: Dict,
    external_inds: Dict,
    data_clamps: Optional[Tuple[str, jnp.ndarray, pd.DataFrame]] = None,
) -> Tuple[Dict, Dict]:
    """Adds clamps to the external inputs.

    Args:
        externals (Dict): Current external inputs.
        external_inds (Dict): Current external indices.
        data_clamps (Optional[Tuple[str, jnp.ndarray, pd.DataFrame]], optional): Additional data clamps. Defaults to None.

    Returns:
        Tuple[Dict, Dict]: Updated external inputs and indices.
    """
    # If a clamp is inserted, add it to the external inputs.
    if data_clamps is not None:
        state_name, clamps, inds = data_clamps
        if state_name in externals.keys():
            externals[state_name] = jnp.concatenate([externals[state_name], clamps])
            external_inds[state_name] = jnp.concatenate(
                [external_inds[state_name], inds.index.to_numpy()]
            )
        else:
            externals[state_name] = clamps
            external_inds[state_name] = inds.index.to_numpy()

    return externals, external_inds


def integrate(
    module: Module,
    params: List[Dict[str, jnp.ndarray]] = [],
    *,
    param_state: Optional[List[Dict]] = None,
    data_stimuli: Optional[Tuple[jnp.ndarray, pd.DataFrame]] = None,
    data_clamps: Optional[Tuple[str, jnp.ndarray, pd.DataFrame]] = None,
    t_max: Optional[float] = None,
    delta_t: float = 0.025,
    solver: str = "bwd_euler",
    voltage_solver: str = "jaxley.stone",
    checkpoint_lengths: Optional[List[int]] = None,
    all_states: Optional[Dict] = None,
    return_states: bool = False,
) -> jnp.ndarray:
    """
    Solves ODE and simulates neuron model.

    Args:
        params: Trainable parameters returned by `get_parameters()`.
        param_state: Parameters returned by `data_set`.
        data_stimuli: Outputs of `.data_stimulate()`, only needed if stimuli change
            across function calls.
        data_clamps: Outputs of `.data_clamp()`, only needed if clamps change across
            function calls.
        t_max: Duration of the simulation in milliseconds. If `t_max` is greater than
            the length of the stimulus input, the stimulus will be padded at the end
            with zeros. If `t_max` is smaller, then the stimulus with be truncated.
        delta_t: Time step of the solver in milliseconds.
        solver: Which ODE solver to use. Either of ["fwd_euler", "bwd_euler",
            "crank_nicolson"].
        tridiag_solver: Algorithm to solve tridiagonal systems. The  different options
            only affect `bwd_euler` and `crank_nicolson` solvers. Either of ["stone",
            "thomas"], where `stone` is much faster on GPU for long branches
            with many compartments and `thomas` is slightly faster on CPU (`thomas` is
            used in NEURON).
        checkpoint_lengths: Number of timesteps at every level of checkpointing. The
            `prod(checkpoint_lengths)` must be larger or equal to the desired number of
            simulated timesteps. Warning: the simulation is run for
            `prod(checkpoint_lengths)` timesteps, and the result is posthoc truncated
            to the desired simulation length. Therefore, a poor choice of
            `checkpoint_lengths` can lead to longer simulation time. If `None`, no
            checkpointing is applied.
        all_states: An optional initial state that was returned by a previous
            `jx.integrate(..., return_states=True)` run. Overrides potentially
            trainable initial states.
        return_states: If True, it returns all states such that the current state of
            the `Module` can be set with `set_states`.
    """

    assert module.initialized, "Module is not initialized, run `._initialize()`."
    module.to_jax()  # Creates `.jaxnodes` from `.nodes` and `.jaxedges` from `.edges`.

    # Initialize the external inputs and their indices.
    externals = module.externals.copy()
    external_inds = module.external_inds.copy()

    # If stimulus is inserted, add it to the external inputs.
    externals, external_inds = add_stimuli(externals, external_inds, data_stimuli)

    # If a clamp is inserted, add it to the external inputs.
    externals, external_inds = add_clamps(externals, external_inds, data_clamps)

    if not externals.keys():
        # No stimulus was inserted and no clamp was set.
        assert (
            t_max is not None
        ), "If no stimulus or clamp are inserted you have to specify the simulation duration at `jx.integrate(..., t_max=)`."

    for key in externals.keys():
        externals[key] = externals[key].T  # Shape `(time, num_stimuli)`.

    if module.recordings.empty:
        raise ValueError("No recordings are set. Please set them.")
    rec_inds = module.recordings.rec_index.to_numpy()
    rec_states = module.recordings.state.to_numpy()

    # Shorten or pad stimulus depending on `t_max`.
    if t_max is not None:
        t_max_steps = int(t_max // delta_t + 1)

        # Pad or truncate the stimulus.
        for key in externals.keys():
            if t_max_steps > externals[key].shape[0]:
                if key == "i":
                    pad = jnp.zeros(
                        (t_max_steps - externals["i"].shape[0], externals["i"].shape[1])
                    )
                    externals["i"] = jnp.concatenate((externals["i"], pad))
                else:
                    raise NotImplementedError(
                        "clamp must be at least as long as simulation."
                    )
            else:
                externals[key] = externals[key][:t_max_steps, :]

    init_fn, step_fn = build_init_and_step_fn(
        module, voltage_solver=voltage_solver, solver=solver
    )
    all_states, all_params = init_fn(params, all_states, param_state, delta_t)

    def _body_fun(state, externals):
        state = step_fn(state, all_params, externals, external_inds, delta_t)
        recs = jnp.asarray(
            [
                state[rec_state][rec_ind]
                for rec_state, rec_ind in zip(rec_states, rec_inds)
            ]
        )
        return state, recs

    # If necessary, pad the stimulus with zeros in order to simulate sufficiently long.
    # The total simulation length will be `prod(checkpoint_lengths)`. At the end, we
    # return only the first `nsteps_to_return` elements (plus the initial state).
    if externals:
        example_key = list(externals.keys())[0]
        nsteps_to_return = len(externals[example_key])
    else:
        nsteps_to_return = t_max_steps

    if checkpoint_lengths is None:
        checkpoint_lengths = [nsteps_to_return]
        length = nsteps_to_return
    else:
        length = prod(checkpoint_lengths)
        size_difference = length - nsteps_to_return
        assert (
            nsteps_to_return <= length
        ), "The desired simulation duration is longer than `prod(nested_length)`."
        if externals:
            dummy_external = jnp.zeros(
                (size_difference, externals[example_key].shape[1])
            )
            for key in externals.keys():
                externals[key] = jnp.concatenate([externals[key], dummy_external])

    # Record the initial state.
    init_recs = jnp.asarray(
        [
            all_states[rec_state][rec_ind]
            for rec_state, rec_ind in zip(rec_states, rec_inds)
        ]
    )
    init_recording = jnp.expand_dims(init_recs, axis=0)

    # Run simulation.
    all_states, recordings = nested_checkpoint_scan(
        _body_fun,
        all_states,
        externals,
        length=length,
        nested_lengths=checkpoint_lengths,
    )
    recs = jnp.concatenate([init_recording, recordings[:nsteps_to_return]], axis=0).T
    return (recs, all_states) if return_states else recs



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, Optional

import jax.numpy as jnp

from jaxley.pumps.pump import Pump


class CaPump(Pump):
    """Calcium dynamics tracking inside calcium concentration

    Modeled after Destexhe et al. 1994.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name)
        self.channel_params = {
            f"{self._name}_gamma": 0.05,  # Fraction of free calcium (not buffered).
            f"{self._name}_decay": 80,  # Buffering time constant in ms.
            f"{self._name}_depth": 0.1,  # Depth of shell in um.
            f"{self._name}_minCaCon_i": 1e-4,  # Minimum intracell. concentration in mM.
        }
        self.channel_states = {"i_Ca": 1e-8, "CaCon_i": 5e-05}
        self.ion_name = "CaCon_i"
        self.current_name = "i_CaPump"
        self.META = {
            "reference": "Modified from Destexhe et al., 1994",
            "mechanism": "Calcium dynamics",
        }

    def update_states(
        self,
        states: Dict[str, jnp.ndarray],
        dt,
        v,
        params: Dict[str, jnp.ndarray],
    ):
        """Update states if necessary (but this pump has no states to update)."""
        return {"CaCon_i": states["CaCon_i"], "i_Ca": states["i_Ca"]}

    def compute_current(
        self,
        states: Dict[str, jnp.ndarray],
        modified_state,
        params: Dict[str, jnp.ndarray],
    ):
        """Return change of calcium concentration based on calcium current and decay."""
        prefix = self._name
        ica = states["i_Ca"]
        gamma = params[f"{prefix}_gamma"]
        decay = params[f"{prefix}_decay"]
        depth = params[f"{prefix}_depth"]
        minCai = params[f"{prefix}_minCaCon_i"]

        FARADAY = 96485  # Coulombs per mole.

        # Calculate the contribution of calcium currents to cai change.
        #
        # Note the similarity to the update in `CaCurrentToConcentrationChange`. The
        # main difference (apart from the multiplication with gamma) is that
        # `CaCurrentToConcentrationChange` multiplies by `surface_area / volume`,
        # whereas this channel multiplies by `1/depth`. However, If one assumes calcium
        # to be stored in a ring of width `depth` just inside the membrane, then A/V is:
        # `2r / (r^2 - (r-depth)^2)`. For any r >> depth, this equation is approximately
        # `1/depth`. It holds quite well as long as `r > 5*depth`.
        drive_channel = -10_000.0 * ica * gamma / (2 * FARADAY * depth)

        # Return to steady state ("leak").
        state_decay = (modified_state - minCai) / decay

        # Total calcium update is the sum of the two.
        diff = drive_channel - state_decay
        return -diff

    def init_state(
        self,
        states: Dict[str, jnp.ndarray],
        v: jnp.ndarray,
        params: Dict[str, jnp.ndarray],
        delta_t: float,
    ):
        """Initialize states of channel."""
        return {}



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"
import numpy as np
import pytest
from neuron import h

import jaxley as jx
from jaxley.channels import HH

_ = h.load_file("stdlib.hoc")
_ = h.load_file("import3d.hoc")


# Test is failing for "morph.swc". This is because NEURON and Jaxley handle interrupted
# soma differently, see issue #140.
@pytest.mark.parametrize("file", ["morph_single_point_soma.swc", "morph_minimal.swc"])
def test_swc_reader_lengths(file, swc2jaxley):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    _, pathlengths, _, _, _ = swc2jaxley(fname, max_branch_len=2000.0)
    if pathlengths[0] == 0.1:
        pathlengths = pathlengths[1:]

    for sec in h.allsec():
        h.delete_section(sec=sec)

    cell = h.Import3d_SWC_read()
    cell.input(fname)
    i3d = h.Import3d_GUI(cell, False)
    i3d.instantiate(None)

    neuron_pathlengths = []
    for sec in h.allsec():
        neuron_pathlengths.append(sec.L)
    neuron_pathlengths = np.asarray(neuron_pathlengths)

    for p in pathlengths:
        dists = np.abs(neuron_pathlengths - p)
        assert np.min(dists) < 1e-3, "Some branches have too large distance."

    assert len(pathlengths) == len(
        neuron_pathlengths
    ), "Number of branches does not match."


def test_dummy_compartment_length(swc2jaxley):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_soma_both_ends.swc")

    parents, pathlengths, _, _, _ = swc2jaxley(fname, max_branch_len=2000.0)
    assert parents == [-1, 0, 0, 1]
    assert pathlengths == [0.1, 1.0, 2.6, 2.2]


@pytest.mark.parametrize("file", ["morph_250_single_point_soma.swc", "morph_250.swc"])
def test_swc_radius(file, swc2jaxley):
    """We expect them to match for sufficiently large ncomp. See #140."""
    ncomp = 64
    non_split = 1 / ncomp
    range_16 = np.linspace(non_split / 2, 1 - non_split / 2, ncomp)

    # Can not use full morphology because of branch sorting.
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    _, pathlen, radius_fns, _, _ = swc2jaxley(fname, max_branch_len=2000.0, sort=False)
    jaxley_diams = []
    for r in radius_fns:
        jaxley_diams.append(r(range_16) * 2)

    for sec in h.allsec():
        h.delete_section(sec=sec)

    cell = h.Import3d_SWC_read()
    cell.input(fname)
    i3d = h.Import3d_GUI(cell, False)
    i3d.instantiate(None)

    neuron_diams = []
    for sec in h.allsec():
        sec.nseg = ncomp
        diams_in_branch = []
        for seg in sec:
            diams_in_branch.append(seg.diam)
        neuron_diams.append(diams_in_branch)
    neuron_diams = np.asarray(neuron_diams)

    for i in range(len(jaxley_diams)):
        max_error = np.max(np.abs(jaxley_diams[i] - neuron_diams[i]))
        assert max_error < 0.5, f"radiuses do not match, error {max_error}."


@pytest.mark.parametrize("file", ["morph_single_point_soma.swc", "morph.swc"])
def test_swc_voltages(file, SimpleMorphCell, swc2jaxley):
    """Check if voltages of SWC recording match.

    To match the branch indices between NEURON and jaxley, we rely on comparing the
    length of the branches.

    It tests whether, on average over time and recordings, the voltage is off by less
    than 1.5 mV.
    """
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)  # n120

    i_delay = 2.0
    i_dur = 5.0
    i_amp = 0.25
    t_max = 20.0
    dt = 0.025

    ncomp_per_branch = 8

    ##################### NEURON ##################
    h.secondorder = 0

    for sec in h.allsec():
        h.delete_section(sec=sec)

    cell = h.Import3d_SWC_read()
    cell.input(fname)
    i3d = h.Import3d_GUI(cell, False)
    i3d.instantiate(None)

    for sec in h.allsec():
        sec.nseg = ncomp_per_branch

    pathlengths_neuron = np.asarray([sec.L for sec in h.allsec()])

    ####################### jaxley ##################
    _, pathlengths, _, _, _ = swc2jaxley(fname, max_branch_len=2_000)
    cell = SimpleMorphCell(fname, ncomp_per_branch, max_branch_len=2_000.0)
    cell.insert(HH())

    trunk_inds = [1, 4, 5, 13, 15, 21, 23, 24, 29, 33]
    tuft_inds = [6, 16, 18, 36, 38, 44, 51, 52, 53, 54]
    basal_inds = np.arange(81, 156, 8).tolist()

    neuron_trunk_inds = []
    for i, p in enumerate(pathlengths):
        if i in trunk_inds:
            closest_match = np.argmin(np.abs(pathlengths_neuron - p))
            neuron_trunk_inds.append(closest_match)

    neuron_tuft_inds = []
    for i, p in enumerate(pathlengths):
        if i in tuft_inds:
            closest_match = np.argmin(np.abs(pathlengths_neuron - p))
            neuron_tuft_inds.append(closest_match)

    neuron_basal_inds = []
    for i, p in enumerate(pathlengths):
        if i in basal_inds:
            closest_match = np.argmin(np.abs(pathlengths_neuron - p))
            neuron_basal_inds.append(closest_match)

    cell.set("axial_resistivity", 1_000.0)
    cell.set("v", -62.0)
    cell.set("HH_m", 0.074901)
    cell.set("HH_h", 0.4889)
    cell.set("HH_n", 0.3644787)

    cell.branch(1).loc(0.05).stimulate(
        jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    )
    for i in trunk_inds + tuft_inds + basal_inds:
        cell.branch(i).loc(0.05).record()

    voltages_jaxley = jx.integrate(cell, delta_t=dt, voltage_solver="jax.sparse")

    ################### NEURON #################
    stim = h.IClamp(h.soma[0](0.1))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    counter = 0
    voltage_recs = {}

    for r in neuron_trunk_inds:
        for i, sec in enumerate(h.allsec()):
            if i == r:
                v = h.Vector()
                v.record(sec(0.05)._ref_v)
                voltage_recs[f"v{counter}"] = v
                counter += 1

    for r in neuron_tuft_inds:
        for i, sec in enumerate(h.allsec()):
            if i == r:
                v = h.Vector()
                v.record(sec(0.05)._ref_v)
                voltage_recs[f"v{counter}"] = v
                counter += 1

    for r in neuron_basal_inds:
        for i, sec in enumerate(h.allsec()):
            if i == r:
                v = h.Vector()
                v.record(sec(0.05)._ref_v)
                voltage_recs[f"v{counter}"] = v
                counter += 1

    for sec in h.allsec():
        sec.insert("hh")
        sec.Ra = 1_000.0

        sec.gnabar_hh = 0.120  # S/cm2
        sec.gkbar_hh = 0.036  # S/cm2
        sec.gl_hh = 0.0003  # S/cm2
        sec.ena = 50  # mV
        sec.ek = -77.0  # mV
        sec.el_hh = -54.3  # mV

    h.dt = dt
    tstop = t_max
    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < tstop:
            h.fadvance()

    initialize()
    integrate()
    voltages_neuron = np.asarray([voltage_recs[key] for key in voltage_recs])
    errors = np.mean(np.abs(voltages_jaxley - voltages_neuron), axis=1)

    ####################### check ################
    assert all(errors < 2.5), "voltages do not match."


@pytest.mark.parametrize(
    "reader_backend",
    [
        "graph",
        "custom",
    ],
)
@pytest.mark.parametrize(
    "file",
    [
        "morph_3_types.swc",
        "morph_3_types_single_point_soma.swc",
        "morph.swc",
        "bbp_with_axon.swc",
    ],
)
def test_swc_types(reader_backend, file):
    # Can not use full morphology because of branch sorting.
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)
    backend_kwargs = (
        {"ignore_swc_trace_errors": False} if reader_backend == "graph" else {}
    )
    cell = jx.read_swc(fname, ncomp=1, backend=reader_backend, **backend_kwargs)
    desired_numbers_of_comps = {
        "morph_3_types.swc": {"soma": 1, "axon": 1, "basal": 1},
        "morph_3_types_single_point_soma.swc": {
            "soma": 1,
            "axon": 1,
            "basal": 1,
        },
        "morph.swc": {"soma": 2, "basal": 101, "apical": 53},
        "bbp_with_axon.swc": {"soma": 1, "axon": 128, "basal": 66, "apical": 129},
    }
    # Test soma.
    for key, n_desired in desired_numbers_of_comps[file].items():
        n_comps_in_morph = len(cell.groups[key])
        assert (
            n_comps_in_morph == n_desired
        ), f"{key} has {n_comps_in_morph} != {n_desired} comps!"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax.numpy as jnp
from jax import tree_util


def l2_norm(x: "PyTree") -> jnp.array:
    """Return the L2-norm of a pytree. Taken from GH jax/issues/3124."""
    leaves, _ = tree_util.tree_flatten(x)
    return jnp.sqrt(sum([jnp.sum(leaf**2) for leaf in leaves]))



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jax import config

from jaxley.connect import connect

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")

import os
from math import pi

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"

import jax.numpy as jnp
import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH
from jaxley.synapses import IonotropicSynapse


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_radius_and_length_compartment(voltage_solver, SimpleComp):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )

    comp = SimpleComp()

    np.random.seed(1)
    comp.set("length", 5 * np.random.rand(1))
    comp.set("radius", np.random.rand(1))

    comp.insert(HH())
    comp.record()
    comp.stimulate(current)

    voltages = jx.integrate(comp, delta_t=dt, voltage_solver=voltage_solver)

    voltages_081123 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703,
                45.83306656,
                29.72581199,
                -15.28462737,
                -39.8274255,
                -66.59985023,
                -75.55397377,
                -75.74933676,
                -75.48829132,
                -75.15524717,
            ]
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_081123))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_radius_and_length_branch(voltage_solver, SimpleBranch):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )

    branch = SimpleBranch(ncomp=2)

    np.random.seed(1)
    branch.set("length", np.flip(5 * np.random.rand(2)))
    branch.set("radius", np.flip(np.random.rand(2)))

    branch.insert(HH())
    branch.loc(0.0).record()
    branch.loc(0.0).stimulate(current)

    voltages = jx.integrate(branch, delta_t=dt, voltage_solver=voltage_solver)

    voltages_300724 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703217397,
                57.69711962162869,
                27.503641673097068,
                -20.26260571077068,
                -44.42160900167978,
                -70.6396420990157,
                -75.76980477606747,
                -75.7453994826729,
                -75.46741856343822,
                -75.1296290199045,
            ]
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_radius_and_length_cell(voltage_solver, SimpleCell):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )

    num_branches = 3
    cell = SimpleCell(num_branches, ncomp=2)

    np.random.seed(1)
    rands1 = 5 * np.random.rand(2 * num_branches)
    rands2 = np.random.rand(2 * num_branches)
    for b in range(num_branches):
        cell.branch(b).set("length", np.flip(rands1[2 * b : 2 * b + 2]))
        cell.branch(b).set("radius", np.flip(rands2[2 * b : 2 * b + 2]))

    cell.insert(HH())
    cell.branch(1).loc(0.0).record()
    cell.branch(1).loc(0.0).stimulate(current)

    voltages = jx.integrate(cell, delta_t=dt, voltage_solver=voltage_solver)

    voltages_300724 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703217234,
                -1.8726881060703575,
                39.98176090770511,
                -2.853384097766068,
                -30.048377518182228,
                -54.259868414343295,
                -73.9359886731972,
                -75.74042745179534,
                -75.5791988501548,
                -75.26829188655702,
            ]
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"


@pytest.mark.parametrize("voltage_solver", ["jaxley.stone", "jax.sparse"])
def test_radius_and_length_net(voltage_solver, SimpleNet):
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.02, delta_t=0.025, t_max=5.0
    )

    num_branches = 3
    net = SimpleNet(2, num_branches, 2)

    np.random.seed(1)
    rands1 = 5 * np.random.rand(2 * num_branches)
    rands2 = np.random.rand(2 * num_branches)
    for b in range(num_branches):
        net.cell(0).branch(b).set("length", np.flip(rands1[2 * b : 2 * b + 2]))
        net.cell(0).branch(b).set("radius", np.flip(rands2[2 * b : 2 * b + 2]))

    np.random.seed(2)
    rands1 = 5 * np.random.rand(2 * num_branches)
    rands2 = np.random.rand(2 * num_branches)
    for b in range(num_branches):
        net.cell(1).branch(b).set("length", np.flip(rands1[2 * b : 2 * b + 2]))
        net.cell(1).branch(b).set("radius", np.flip(rands2[2 * b : 2 * b + 2]))

    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(0).loc(0.0),
        IonotropicSynapse(),
    )
    net.insert(HH())

    # first cell, 0-eth branch, 0-st compartment because loc=0.0
    radius_post = net[1, 0, 0].nodes["radius"].item()
    lenght_post = net[1, 0, 0].nodes["length"].item()
    area = 2 * pi * lenght_post * radius_post
    point_process_to_dist_factor = 100_000.0 / area
    net.set("IonotropicSynapse_gS", 0.5 / point_process_to_dist_factor)

    for cell_ind in range(2):
        net.cell(cell_ind).branch(1).loc(0.0).record()

    for stim_ind in range(2):
        net.cell(stim_ind).branch(1).loc(0.0).stimulate(current)

    voltages = jx.integrate(net, delta_t=dt, voltage_solver=voltage_solver)

    voltages_300724 = jnp.asarray(
        [
            [
                -70.0,
                -66.53085703217131,
                -1.8726881060701073,
                39.98176090770495,
                -2.8533840977659226,
                -30.04837751818565,
                -54.25986841434464,
                -73.93598867319699,
                -75.74042745179538,
                -75.57919885015413,
                -75.2682918865562,
            ],
            [
                -70.0,
                -66.47133094286725,
                -22.42632884453572,
                42.81477131595844,
                5.7572444690008036,
                -23.81466327013234,
                -47.82796157947444,
                -71.80106951477758,
                -75.57738636633145,
                -75.49698994483342,
                -75.17441392712844,
            ],
        ]
    )
    max_error = np.max(np.abs(voltages[:, ::20] - voltages_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np
from jax import jit

import jaxley as jx
from jaxley.channels import HH


def test_constant_and_data_stimulus(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.branch(0).loc(0.0).record("v")

    # test data_stimulate and jit works with trainable parameters see #467
    cell.make_trainable("radius")

    i_amp_const = 0.1
    i_amps_data = jnp.asarray([0.01, 0.005])

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=i_amp_const, delta_t=0.025, t_max=5.0
    )
    cell.branch(1).loc(0.6).stimulate(current)

    def provide_data(i_amps):
        current = jx.datapoint_to_step_currents(
            i_delay=0.5, i_dur=1.0, i_amp=i_amps, delta_t=0.025, t_max=5.0
        )
        data_stimuli = None
        data_stimuli = cell.branch(1).loc(0.6).data_stimulate(current[0], data_stimuli)
        data_stimuli = cell.branch(1).loc(0.6).data_stimulate(current[1], data_stimuli)
        return data_stimuli

    def simulate(i_amps):
        data_stimuli = provide_data(i_amps)
        return jx.integrate(cell, data_stimuli=data_stimuli)

    jitted_simulate = jit(simulate)
    v_data = jitted_simulate(i_amps_data)

    cell.delete_stimuli()
    i_amp_summed = i_amp_const + jnp.sum(i_amps_data)
    current_sum = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=i_amp_summed, delta_t=0.025, t_max=5.0
    )
    cell.branch(1).loc(0.6).stimulate(current_sum)

    v_stim = jx.integrate(cell)

    diff = np.abs(v_stim - v_data)
    assert np.max(diff) < 1e-8


def test_data_vs_constant_stimulus(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.branch(0).loc(0.0).record("v")

    i_amps_data = jnp.asarray([0.01, 0.005])

    def provide_data(i_amps):
        current = jx.datapoint_to_step_currents(0.5, 1.0, i_amps, 0.025, 5.0)
        data_stimuli = None
        data_stimuli = cell.branch(1).loc(0.6).data_stimulate(current[0], data_stimuli)
        data_stimuli = cell.branch(1).loc(0.6).data_stimulate(current[1], data_stimuli)
        return data_stimuli

    def simulate(i_amps):
        data_stimuli = provide_data(i_amps)
        return jx.integrate(cell, data_stimuli=data_stimuli)

    jitted_simulate = jit(simulate)
    v_data = jitted_simulate(i_amps_data)

    cell.delete_stimuli()
    i_amp_summed = jnp.sum(i_amps_data)
    current_sum = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=i_amp_summed, delta_t=0.025, t_max=5.0
    )
    cell.branch(1).loc(0.6).stimulate(current_sum)

    v_stim = jx.integrate(cell)

    diff = np.abs(v_stim - v_data)
    assert np.max(diff) < 1e-8



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")


import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".4"

import jax.numpy as jnp
import numpy as np
import pytest
from jaxley_mech.channels.l5pc import CaHVA
from neuron import h, rxd

import jaxley as jx
from jaxley.channels import HH
from jaxley.pumps import CaFaradayConcentrationChange, CaNernstReversal

_ = h.load_file("stdlib.hoc")
_ = h.load_file("import3d.hoc")


@pytest.mark.parametrize("solver", ["bwd_euler", "crank_nicolson"])
def test_similarity(solver):
    """Test similarity of jaxley vs neuron for a branch.

    The branch has an uneven radius.
    """
    i_delay = 3.0  # ms
    i_dur = 2.0  # ms
    i_amp = 1.0  # nA

    dt = 0.025  # ms
    t_max = 10.0  # ms

    voltages_jaxley = _run_jaxley(i_delay, i_dur, i_amp, dt, t_max, solver)
    voltages_neuron = _run_neuron(i_delay, i_dur, i_amp, dt, t_max, solver)

    assert np.mean(np.abs(voltages_jaxley - voltages_neuron)) < 0.05


def _run_jaxley(i_delay, i_dur, i_amp, dt, t_max, solver):
    ncomp_per_branch = 8
    comp = jx.Compartment()
    branch = jx.Branch([comp for _ in range(ncomp_per_branch)])
    branch.insert(HH())

    radiuses = np.linspace(3.0, 15.0, ncomp_per_branch)
    for i, loc in enumerate(np.linspace(0, 1, ncomp_per_branch)):
        branch.loc(loc).set("radius", radiuses[i])

    branch.set("length", 10.0)
    branch.set("axial_resistivity", 1_000.0)
    branch.set("capacitance", 5.0)

    branch.set("HH_gNa", 0.120)
    branch.set("HH_gK", 0.036)
    branch.set("HH_gLeak", 0.0003)
    branch.set("HH_m", 0.07490098835688629)
    branch.set("HH_h", 0.488947681848153)
    branch.set("HH_n", 0.3644787002343737)
    branch.set("v", -62.0)

    current = jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    branch.loc(0.0).stimulate(current)
    branch.loc(0.0).record()
    branch.loc(1.0).record()

    voltages = jx.integrate(branch, delta_t=dt, solver=solver)

    return voltages


def _run_neuron(i_delay, i_dur, i_amp, dt, t_max, solver):
    if solver == "bwd_euler":
        h.secondorder = 0
    elif solver == "crank_nicolson":
        h.secondorder = 1
    else:
        raise ValueError

    ncomp_per_branch = 8
    h.dt = dt

    for sec in h.allsec():
        h.delete_section(sec=sec)
    branch = h.Section()

    branch.nseg = ncomp_per_branch
    branch.Ra = 1_000.0
    branch.L = 10.0 * ncomp_per_branch
    branch.cm = 5.0

    radiuses = np.linspace(3.0, 15.0, ncomp_per_branch)
    for i, comp in enumerate(branch):
        comp.diam = 2 * radiuses[i]

    branch.insert("hh")
    branch.gnabar_hh = 0.120  # S/cm2
    branch.gkbar_hh = 0.036  # S/cm2
    branch.gl_hh = 0.0003  # S/cm2
    branch.ena = 50  # mV
    branch.ek = -77.0  # mV
    branch.el_hh = -54.3  # mV

    stim = h.IClamp(branch(0.05))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    voltage1 = h.Vector()
    voltage1.record(branch(0.05)._ref_v)
    voltage2 = h.Vector()
    voltage2.record(branch(0.95)._ref_v)

    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < t_max:
            h.fadvance()

    initialize()
    integrate()

    voltages = np.asarray([list(voltage1), list(voltage2)])
    return voltages


@pytest.mark.parametrize("solver", ["bwd_euler", "crank_nicolson"])
def test_similarity_complex(solver):
    """Test for a branch where radius varies for every seg and l and r_a varies for
    two sub-branches.

    Since NEURON enforces that all segments within a branch have the same length and
    the same axial resistivity, we have to create two NEURON branches. In jaxley,
    we only need one branch.
    """
    i_delay = 2.0
    i_dur = 5.0
    i_amp = 0.1
    t_max = 20.0
    dt = 0.025

    diams = [
        15.925172886886058,
        15.299516921156213,
        14.662901345202302,
        13.471623855881704,
        12.095324558467158,
        9.602043499254922,
        8.155112133121278,
        6.48666134956395,
        3.873367784857285,
        2.729358541238842,
        2.14754416410492,
        2.064491286216302,
        1.6082726642214933,
        1.2703115584172973,
        0.9684275792140471,
        0.8000000119209283,
    ]
    capacitances = np.linspace(1.0, 10.0, 16)
    voltages_jaxley = _jaxley_complex(
        i_delay, i_dur, i_amp, dt, t_max, diams, capacitances, solver
    )
    voltages_neuron = _neuron_complex(
        i_delay, i_dur, i_amp, dt, t_max, diams, capacitances, solver
    )

    assert np.mean(np.abs(voltages_jaxley - voltages_neuron)) < 0.05


def _jaxley_complex(i_delay, i_dur, i_amp, dt, t_max, diams, capacitances, solver):
    ncomp = 16
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp)

    branch.insert(HH())

    branch.set("v", -62.0)
    branch.set("HH_m", 0.074901)
    branch.set("HH_h", 0.4889)
    branch.set("HH_n", 0.3644787)

    for loc in np.linspace(0, 0.45, 8):
        branch.loc(loc).set("length", 20.0)

    for loc in np.linspace(0.55, 1.0, 8):
        branch.loc(loc).set("length", 100.0)

    for loc in np.linspace(0, 0.45, 8):
        branch.loc(loc).set("axial_resistivity", 800.0)

    for loc in np.linspace(0.55, 1.0, 8):
        branch.loc(loc).set("axial_resistivity", 800.0)

    counter = 0
    for loc in np.linspace(0, 1, ncomp):
        branch.loc(loc).set("radius", diams[counter] / 2)
        branch.loc(loc).set("capacitance", capacitances[counter])
        counter += 1

    # 0.02 is fine here because ncomp=8 for NEURON, but ncomp=16 for jaxley.
    current = jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    branch.loc(0.02).stimulate(current)
    branch.loc(0.02).record()
    branch.loc(0.52).record()
    branch.loc(0.98).record()

    s = jx.integrate(branch, delta_t=dt, solver=solver)
    return s


def _neuron_complex(i_delay, i_dur, i_amp, dt, t_max, diams, capacitances, solver):
    if solver == "bwd_euler":
        h.secondorder = 0
    elif solver == "crank_nicolson":
        h.secondorder = 1
    else:
        raise ValueError

    for sec in h.allsec():
        h.delete_section(sec=sec)

    branch1 = h.Section()
    branch2 = h.Section()

    branch2.connect(branch1, 1, 0)

    branch1.nseg = 8
    branch1.L = 160.0
    branch1.Ra = 800.0

    branch2.nseg = 8
    branch2.L = 800.0
    branch2.Ra = 800.0

    counter = 0
    for sec in h.allsec():
        sec.insert("hh")
        sec.gnabar_hh = 0.120  # S/cm2
        sec.gkbar_hh = 0.036  # S/cm2
        sec.gl_hh = 0.0003  # S/cm2
        sec.ena = 50  # mV
        sec.ek = -77.0  # mV
        sec.el_hh = -54.3  # mV

        for i, seg in enumerate(sec):
            seg.diam = diams[counter]
            seg.cm = capacitances[counter]
            counter += 1

    # 0.05 is fine here because ncomp=8, but ncomp=16 for jaxley.
    stim = h.IClamp(branch1(0.05))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    # 0.05 is fine here because ncomp=8, but ncomp=16 for jaxley.
    voltage_recs = {}
    v = h.Vector()
    v.record(branch1(0.05)._ref_v)
    voltage_recs["v0"] = v

    v = h.Vector()
    v.record(branch2(0.05)._ref_v)
    voltage_recs["v1"] = v

    v = h.Vector()
    v.record(branch2(0.95)._ref_v)
    voltage_recs["v2"] = v

    h.dt = dt
    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < t_max:
            h.fadvance()

    initialize()
    integrate()

    return np.asarray([voltage_recs[key] for key in voltage_recs])


def _jaxley_clamped(dt, t_max, diams, v_init, length):
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=2)

    branch.comp(0).set("radius", diams[0] / 2)
    branch.comp(1).set("radius", diams[1] / 2)

    branch.set("axial_resistivity", 100_000.0)
    branch.set("length", length / 2)
    branch.set("v", v_init)
    branch.insert(CaHVA())
    branch.insert(CaFaradayConcentrationChange())
    branch.insert(CaNernstReversal())
    branch.comp(0).set("CaHVA_gCaHVA", 0.00001)
    branch.comp(1).set("CaHVA_gCaHVA", 0.0)
    branch.set("eCa", 0.0)
    branch.comp(0).clamp("v", jnp.linspace(-65.0, 0.0, 40_001))
    branch.init_states()

    branch.comp(0).record("v")
    branch.comp(1).record("v")

    branch.comp(0).record("CaCon_i")
    branch.comp(1).record("CaCon_i")

    voltages = jx.integrate(branch, t_max=t_max, delta_t=dt)
    return voltages


def _neuron_clamped(dt, t_max, diams, v_init, length):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "arm64/.libs/libnrnmech.so")
    h.nrn_load_dll(fname)

    h.secondorder = 0
    h.dt = dt

    for sec in h.allsec():
        h.delete_section(sec=sec)

    branch1 = h.Section()

    branch1.nseg = 2
    branch1.Ra = 100_000.0
    branch1.L = length
    branch1.insert("Ca_HVA")
    for i, seg in enumerate(branch1):
        if i == 0:
            seg.gCa_HVAbar_Ca_HVA = 0.00001
        else:
            seg.gCa_HVAbar_Ca_HVA = 0.0

    for i, seg in enumerate(branch1):
        seg.diam = diams[i]

    # Set diffusion to zero. This is causing the different interpretation of `.area()`
    # of NEURON and therefore causes this test to fail (also for d>0) if the radiuses
    # of compartments are not the same (as is the case here).
    r = rxd.Region(h.allsec(), nrn_region="i", geometry=None)
    ca = rxd.Species(r, d=0, name="ca", charge=2)

    # Voltage clamp.
    t = h.Vector(np.arange(0, 1000, dt))
    y = h.Vector(np.linspace(-65.0, 0.0, len(t)))
    y.play(branch1(0.3)._ref_v, t, True)

    v1 = h.Vector()
    v1.record(branch1(0.3)._ref_v)

    v2 = h.Vector()
    v2.record(branch1(0.7)._ref_v)

    cacon1 = h.Vector()
    cacon1.record(branch1(0.3)._ref_cai)

    cacon2 = h.Vector()
    cacon2.record(branch1(0.7)._ref_cai)

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < t_max:
            h.fadvance()

    initialize()
    integrate()

    v1 = np.asarray(list(v1))
    v2 = np.asarray(list(v2))
    cacon1 = np.asarray(list(cacon1))
    cacon2 = np.asarray(list(cacon2))
    return np.stack([v1, v2, cacon1, cacon2])


@pytest.mark.xfail
def test_similarity_voltage_clamp_cacon_i_behavior():
    """Test whether two compartments were one is clamped gives same results.

    In this test, we clamp the voltage of one of the compartments. The rational is that
    this decouples any influence of _voltage_ diffusion on the calcium dynamics (in
    comp 0). This allowed us to study why the calcium dynamics between NEURON and
    Jaxley differ for comp(0) of a branch, even when the voltages are the same and
    diffusion is d=0.

    This test currently fails. The axial conductance is defined as:
    ```1 / R_long / area_of_sink_compartment```
    However, NEURON computes `area_of_sink_compartment` differently _depending on
    whether diffusion is activated or not_. This make no sense to me (@michaeldeistler),
    so I am not implementing it in Jaxley.

    Jaxley simply assumes a cylindrical compartment: `A = 2 * pi * r * l`
    Without diffusion, NEURON does the same. With diffusion, however, NEURON creates
    frustums: `A = 2 * pi * r * (l/2) + frustum_area`.
    The frustum radiuses are on one side, the radius of the compartment, and on the
    other side the radius at the connection point (which itself is the weighted
    average of all neighboring compartments). Because of this, the radius of `comp(1)`
    influences the `.area()` of `comp(0)`, and thus impacts its dynamics.

    Notably, NEURON even uses this updated area (which can be inspected with
    `branch(0.3).area()` _after having run the simulation with diffusion at least
    once_) even for the voltage equations. See also #140.
    """
    t_max = 1_000.01
    dt = 0.025

    diams = [4.0, 8.0]  # The test passes when diams has the same values.
    length = 100.0
    v_init = -65.0

    v_and_cacon_jaxley = _jaxley_clamped(dt, t_max, diams, v_init, length)
    v_and_cacon_neuron = _neuron_clamped(dt, t_max, diams, v_init, length)
    v_jaxley = v_and_cacon_jaxley[:2]
    cacon_jaxley = v_and_cacon_jaxley[2:]
    v_neuron = v_and_cacon_neuron[:2]
    cacon_neuron = v_and_cacon_neuron[2:]

    v_max_error = np.max(np.abs(v_jaxley - v_neuron))
    cacon_max_error = np.max(np.abs(cacon_jaxley - cacon_neuron))
    # Note: for very thin cables (<1um radius), this test can fail by a tiny bit, even
    # if both comps have the same diameter.
    assert cacon_max_error < 2e-6, f"CaCon_i error {cacon_max_error}"

    # To be on the safe side, we also check voltages. However, this test is generally
    # about CaCon_i.
    assert v_max_error < 1e-2, f"Voltage error {v_max_error}"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os
import re
from pathlib import Path
from typing import List

import numpy as np


def test_rm_all_deprecated_functions():
    from jaxley.__version__ import __version__ as package_version

    package_version = np.array([int(s) for s in package_version.split(".")])

    # Pattern to match both @deprecated and @deprecated_kwargs
    decorator_pattern = r"@deprecated(?:_kwargs)?"
    version_pattern = r"[v]?(\d+\.\d+\.\d+)"

    package_dir = Path(__file__).parent.parent / "jaxley"
    project_root = Path(__file__).parent.parent

    violations = []
    for py_file in package_dir.rglob("*.py"):
        with open(py_file, "r") as f:
            for line_num, line in enumerate(f, 1):
                if re.search(decorator_pattern, line):
                    version_match = re.search(version_pattern, line)
                    if version_match:
                        depr_version_str = version_match.group(1)
                        depr_version = np.array(
                            [int(s) for s in depr_version_str.split(".")]
                        )
                        if np.all(depr_version <= package_version):
                            relative_path = py_file.relative_to(project_root)
                            violations.append(f"{relative_path}:L{line_num}")

    assert not violations, "\n".join(
        ["Found deprecated items that should have been removed:", *violations]
    )



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os
from copy import deepcopy

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")


os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"

import jax.numpy as jnp
import networkx as nx
import numpy as np
import pandas as pd
import pytest

import jaxley as jx
from jaxley import connect
from jaxley.channels import HH
from jaxley.channels.pospischil import K, Leak, Na
from jaxley.io.graph import (
    add_missing_graph_attrs,
    from_graph,
    make_jaxley_compatible,
    swc_to_graph,
    to_graph,
    trace_branches,
)
from jaxley.synapses import IonotropicSynapse, TestSynapse

# from jaxley.utils.misc_utils import recursive_compare
from tests.helpers import (
    equal_both_nan_or_empty_df,
    get_segment_xyzrL,
    import_neuron_morph,
    jaxley2neuron_by_group,
    match_stim_loc,
)


# test exporting and re-importing of different modules
def test_graph_import_export_cycle(
    SimpleComp, SimpleBranch, SimpleCell, SimpleNet, SimpleMorphCell
):
    np.random.seed(0)
    comp = SimpleComp()
    branch = SimpleBranch(4)
    cell = SimpleCell(5, 4)
    morph_cell = SimpleMorphCell(ncomp=1)
    net = SimpleNet(3, 5, 4)

    # add synapses
    connect(net[0, 0, 0], net[1, 0, 0], IonotropicSynapse())
    connect(net[0, 0, 1], net[1, 0, 1], IonotropicSynapse())
    # connect(net[0, 0, 1], net[1, 0, 1], TestSynapse()) # makes test fail, see warning w. synapses = True

    # add groups
    net.cell(2).add_to_group("cell2")
    net.cell(2).branch(1).add_to_group("cell2branch1")

    # add ion channels
    net.cell(0).insert(Na())
    net.cell(0).insert(Leak())
    net.cell(1).branch(1).insert(Na())
    net.cell(0).insert(K())

    # test consistency of exported and re-imported modules
    for module in [comp, branch, cell, net, morph_cell]:
        module.compute_xyz()  # ensure x,y,z in nodes b4 exporting for later comparison
        module_graph = to_graph(
            module, channels=True, synapses=True
        )  # ensure to_graph works
        re_module = from_graph(module_graph)  # ensure prev exported graph can be read
        re_module_graph = to_graph(
            re_module, channels=True, synapses=True
        )  # ensure to_graph works for re-imported modules

        # ensure original module and re-imported module are equal
        assert np.all(equal_both_nan_or_empty_df(re_module.nodes, module.nodes))
        assert np.all(equal_both_nan_or_empty_df(re_module.edges, module.edges))
        assert np.all(
            equal_both_nan_or_empty_df(re_module.branch_edges, module.branch_edges)
        )

        for k in module.groups:
            assert k in re_module.groups
            assert np.all(re_module.groups[k] == module.groups[k])

        for re_xyzr, xyzr in zip(re_module.xyzr, module.xyzr):
            re_xyzr[np.isnan(re_xyzr)] = -1
            xyzr[np.isnan(xyzr)] = -1

            assert np.all(re_xyzr == xyzr)

        re_imported_mechs = re_module.channels + re_module.synapses
        for re_mech, mech in zip(re_imported_mechs, module.channels + module.synapses):
            assert np.all(re_mech.name == mech.name)

        # ensure exported graph and re-exported graph are equal
        node_df = pd.DataFrame(
            [d for i, d in module_graph.nodes(data=True)], index=module_graph.nodes
        ).sort_index()
        re_node_df = pd.DataFrame(
            [d for i, d in re_module_graph.nodes(data=True)],
            index=re_module_graph.nodes,
        ).sort_index()
        assert np.all(equal_both_nan_or_empty_df(node_df, re_node_df))

        edges = pd.DataFrame(
            [
                {
                    "pre_global_comp_index": i,
                    "post_global_comp_index": j,
                    **module_graph.edges[i, j],
                }
                for (i, j) in module_graph.edges
            ]
        )
        re_edges = pd.DataFrame(
            [
                {
                    "pre_global_comp_index": i,
                    "post_global_comp_index": j,
                    **re_module_graph.edges[i, j],
                }
                for (i, j) in re_module_graph.edges
            ]
        )
        assert np.all(equal_both_nan_or_empty_df(edges, re_edges))

        # ignore "externals", "recordings", "trainable_params", "indices_set_by_trainables"
        for k in ["ncomp", "xyzr"]:
            assert module_graph.graph[k] == re_module_graph.graph[k]

        # assume if module can be integrated, so can be comp, cell and branch
        if isinstance(module, jx.Network):
            # test integration of re-imported module
            re_module.select(nodes=0).record(verbose=False)
            jx.integrate(re_module, t_max=0.5)


@pytest.mark.parametrize(
    "file", ["morph_single_point_soma.swc", "morph.swc", "bbp_with_axon.swc"]
)
def test_trace_branches(file):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)
    graph = swc_to_graph(fname)

    # pre-processing
    graph = add_missing_graph_attrs(graph)
    graph = trace_branches(graph, None, ignore_swc_trace_errors=False)

    edges = pd.DataFrame([{"u": u, "v": v, **d} for u, v, d in graph.edges(data=True)])
    nx_branch_lens = edges.groupby("branch_index")["l"].sum().to_numpy()
    nx_branch_lens = np.sort(nx_branch_lens)

    # exclude artificial root branch
    if np.isclose(nx_branch_lens[0], 1e-1):
        nx_branch_lens = nx_branch_lens[1:]

    h, _ = import_neuron_morph(fname)
    neuron_branch_lens = np.sort([sec.L for sec in h.allsec()])

    errors = np.abs(neuron_branch_lens - nx_branch_lens)
    # one error is expected, see https://github.com/jaxleyverse/jaxley/issues/140
    assert sum(errors > 1e-3) <= 1


@pytest.mark.parametrize(
    "file", ["morph_single_point_soma.swc", "morph.swc", "bbp_with_axon.swc"]
)
def test_from_graph_vs_NEURON(file):
    ncomp = 8
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    graph = swc_to_graph(fname)
    cell = from_graph(
        graph, ncomp=ncomp, max_branch_len=2000, ignore_swc_trace_errors=False
    )
    cell.compute_compartment_centers()
    h, neuron_cell = import_neuron_morph(fname, ncomp=ncomp)

    # remove root branch
    jaxley_comps = cell.nodes[
        ~np.isclose(cell.nodes["length"], 0.1 / ncomp)
    ].reset_index(drop=True)

    jx_branch_lens = (
        jaxley_comps.groupby("global_branch_index")["length"].sum().to_numpy()
    )

    # match by branch lengths
    neuron_xyzd = [np.array(s.psection()["morphology"]["pts3d"]) for s in h.allsec()]
    neuron_branch_lens = np.array(
        [
            np.sqrt((np.diff(n[:, :3], axis=0) ** 2).sum(axis=1)).sum()
            for n in neuron_xyzd
        ]
    )
    neuron_inds = np.argsort(neuron_branch_lens)
    jx_inds = np.argsort(jx_branch_lens)

    neuron_df = pd.DataFrame(columns=["neuron_idx", "x", "y", "z", "radius", "length"])
    jx_df = pd.DataFrame(columns=["jx_idx", "x", "y", "z", "radius", "length"])
    for k in range(len(neuron_inds)):
        neuron_comp_k = np.array(
            [
                get_segment_xyzrL(list(h.allsec())[neuron_inds[k]], comp_idx=i)
                for i in range(ncomp)
            ]
        )
        # make this a dataframe
        neuron_comp_k = pd.DataFrame(
            neuron_comp_k, columns=["x", "y", "z", "radius", "length"]
        )
        neuron_comp_k["idx"] = neuron_inds[k]
        jx_comp_k = jaxley_comps[jaxley_comps["global_branch_index"] == jx_inds[k]][
            ["x", "y", "z", "radius", "length"]
        ]
        jx_comp_k["idx"] = jx_inds[k]
        neuron_df = pd.concat([neuron_df, neuron_comp_k], axis=0, ignore_index=True)
        jx_df = pd.concat([jx_df, jx_comp_k], axis=0, ignore_index=True)

    errors = neuron_df["neuron_idx"].to_frame()
    errors["jx_idx"] = jx_df["jx_idx"]
    errors[["x", "y", "z"]] = neuron_df[["x", "y", "z"]] - jx_df[["x", "y", "z"]]
    errors["xyz"] = np.sqrt((errors[["x", "y", "z"]] ** 2).sum(axis=1))
    errors["radius"] = neuron_df["radius"] - jx_df["radius"]
    errors["length"] = neuron_df["length"] - jx_df["length"]

    # one error is expected, see https://github.com/jaxleyverse/jaxley/issues/140
    assert sum(errors.groupby("jx_idx")["xyz"].max() > 1e-3) <= 1
    assert sum(errors.groupby("jx_idx")["radius"].max() > 1e-3) <= 1
    assert sum(errors.groupby("jx_idx")["length"].max() > 1e-3) <= 1


def test_edges_only_to_jaxley():
    # test if edge graph can pe imported into to jaxley
    sets_of_edges = [
        [(0, 1), (1, 2), (2, 3)],
        [(0, 1), (1, 2), (1, 3), (2, 4), (2, 5), (3, 6), (3, 7)],
    ]
    for edges in sets_of_edges:
        edge_graph = nx.DiGraph(edges)
        edge_module = from_graph(edge_graph)


@pytest.mark.slow
@pytest.mark.parametrize("file", ["morph_single_point_soma.swc", "morph.swc"])
def test_swc2graph_voltages(file):
    """Check if voltages of SWC recording match.

    To match the branch indices between NEURON and jaxley, we rely on comparing the
    length of the branches.

    It tests whether, on average over time and recordings, the voltage is off by less
    than 1.5 mV.
    """
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)  # n120

    ncomp = 8

    i_delay = 2.0
    i_dur = 5.0
    i_amp = 0.25
    t_max = 20.0
    dt = 0.025

    ##################### NEURON ##################
    h, neuron_cell = import_neuron_morph(fname, ncomp=ncomp)

    ####################### jaxley ##################
    graph = swc_to_graph(fname)
    jx_cell = from_graph(
        graph, ncomp=ncomp, max_branch_len=2000, ignore_swc_trace_errors=False
    )
    jx_cell.compute_compartment_centers()
    jx_cell.insert(HH())

    branch_loc = 0.05
    neuron_inds, jaxley_inds = jaxley2neuron_by_group(
        jx_cell, h.allsec(), loc=branch_loc
    )
    trunk_inds, tuft_inds, basal_inds = [
        jaxley_inds[key] for key in ["trunk", "tuft", "basal"]
    ]
    neuron_trunk_inds, neuron_tuft_inds, neuron_basal_inds = [
        neuron_inds[key] for key in ["trunk", "tuft", "basal"]
    ]

    stim_loc = 0.1
    stim_idx = match_stim_loc(jx_cell, h.soma[0], loc=stim_loc)

    jx_cell.set("axial_resistivity", 1_000.0)
    jx_cell.set("v", -62.0)
    jx_cell.set("HH_m", 0.074901)
    jx_cell.set("HH_h", 0.4889)
    jx_cell.set("HH_n", 0.3644787)

    jx_cell.select(stim_idx).stimulate(
        jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    )
    for i in trunk_inds + tuft_inds + basal_inds:
        jx_cell.branch(i).loc(branch_loc).record()

    voltages_jaxley = jx.integrate(jx_cell, delta_t=dt)

    ################### NEURON #################
    stim = h.IClamp(h.soma[0](stim_loc))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    counter = 0
    voltage_recs = {}

    for r in neuron_trunk_inds:
        for i, sec in enumerate(h.allsec()):
            if i == r:
                v = h.Vector()
                v.record(sec(branch_loc)._ref_v)
                voltage_recs[f"v{counter}"] = v
                counter += 1

    for r in neuron_tuft_inds:
        for i, sec in enumerate(h.allsec()):
            if i == r:
                v = h.Vector()
                v.record(sec(branch_loc)._ref_v)
                voltage_recs[f"v{counter}"] = v
                counter += 1

    for r in neuron_basal_inds:
        for i, sec in enumerate(h.allsec()):
            if i == r:
                v = h.Vector()
                v.record(sec(branch_loc)._ref_v)
                voltage_recs[f"v{counter}"] = v
                counter += 1

    for sec in h.allsec():
        sec.insert("hh")
        sec.Ra = 1_000.0

        sec.gnabar_hh = 0.120  # S/cm2
        sec.gkbar_hh = 0.036  # S/cm2
        sec.gl_hh = 0.0003  # S/cm2
        sec.ena = 50  # mV
        sec.ek = -77.0  # mV
        sec.el_hh = -54.3  # mV

    h.dt = dt
    tstop = t_max
    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < tstop:
            h.fadvance()

    initialize()
    integrate()
    voltages_neuron = np.asarray([voltage_recs[key] for key in voltage_recs])

    ####################### check ################
    errors = np.mean(np.abs(voltages_jaxley - voltages_neuron), axis=1)

    assert all(errors < 2.5), "voltages do not match."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

# This folder checks whether the new version of the code produces **exactly** the same
# simulation results as previous versions.



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jaxley.channels.channel import Channel  # isort: skip
from jaxley.channels.hh import HH
from jaxley.channels.pospischil import CaL, CaT, K, Km, Leak, Na
